import os.path

from six.moves import configparser
from zope.interface import implementer

from ldaptor import interfaces
from ldaptor.protocols.ldap import distinguishedname


class MissingBaseDNError(Exception):
    """Configuration must specify a base DN"""

    def __str__(self):
        return self.__doc__


@implementer(interfaces.ILDAPConfig)
class LDAPConfig(object):

    baseDN = None
    identityBaseDN = None
    identitySearch = None

    def __init__(self,
                 baseDN=None,
                 serviceLocationOverrides=None,
                 identityBaseDN=None,
                 identitySearch=None):
        if baseDN is not None:
            baseDN = distinguishedname.DistinguishedName(baseDN)
            self.baseDN = baseDN
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k,v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn] = v
        if identityBaseDN is not None:
            identityBaseDN = distinguishedname.DistinguishedName(identityBaseDN)
            self.identityBaseDN = identityBaseDN
        if identitySearch is not None:
            self.identitySearch = identitySearch

    def getBaseDN(self):
        if self.baseDN is not None:
            return self.baseDN

        cfg = loadConfig()
        try:
            return cfg.get('ldap', 'base')
        except (configparser.NoOptionError,
                configparser.NoSectionError):
            raise MissingBaseDNError()

    def getServiceLocationOverrides(self):
        r = self._loadServiceLocationOverrides()
        r.update(self.serviceLocationOverrides)
        return r

    def _loadServiceLocationOverrides(self):
        serviceLocationOverride = {}
        cfg = loadConfig()
        for section in cfg.sections():
            if section.lower().startswith('service-location '):
                base = section[len('service-location '):].strip()

                host = None
                if cfg.has_option(section, 'host'):
                    host = cfg.get(section, 'host')
                    if not host:
                        host = None

                port = None
                if cfg.has_option(section, 'port'):
                    port = cfg.get(section, 'port')
                    if not port:
                        port = None

                dn = distinguishedname.DistinguishedName(stringValue=base)
                serviceLocationOverride[dn] = (host, port)
        return serviceLocationOverride

    def copy(self, **kw):
        if 'baseDN' not in kw:
            kw['baseDN'] = self.baseDN
        if 'serviceLocationOverrides' not in kw:
            kw['serviceLocationOverrides'] = self.serviceLocationOverrides
        if 'identityBaseDN' not in kw:
            kw['identityBaseDN'] = self.identityBaseDN
        if 'identitySearch' not in kw:
            kw['identitySearch'] = self.identitySearch
        r = self.__class__(**kw)
        return r

    def getIdentityBaseDN(self):
        if self.identityBaseDN is not None:
            return self.identityBaseDN

        cfg = loadConfig()
        try:
            return cfg.get('authentication', 'identity-base')
        except (configparser.NoOptionError,
                configparser.NoSectionError):
            return self.getBaseDN()

    def getIdentitySearch(self, name):
        data = {
            'name': name,
            }

        if self.identitySearch is not None:
            f = self.identitySearch % data
        else:
            cfg = loadConfig()
            try:
                f = cfg.get('authentication', 'identity-search', vars=data)
            except (configparser.NoOptionError,
                    configparser.NoSectionError):
                f = '(|(cn=%(name)s)(uid=%(name)s))' % data
        return f


DEFAULTS = {
    'samba': {'use-lmhash': 'no'},
}

CONFIG_FILES = [
    '/etc/ldaptor/global.cfg',
    os.path.expanduser('~/.ldaptor/global.cfg'),
    ]

__config = None


def loadConfig(configFiles=None,
               reload=False):
    """
    Load configuration file.
    """
    global __config
    if __config is None or reload:
        x = configparser.SafeConfigParser()

        for section, options in DEFAULTS.items():
            x.add_section(section)
            for option, value in options.items():
                x.set(section, option, value)

        if configFiles is None:
            configFiles = CONFIG_FILES
        x.read(configFiles)
        __config = x
    return __config


def useLMhash():
    """
    Read configuration file if necessary and return whether
    to use LanMan hashes or not.
    """
    cfg = loadConfig()
    return cfg.getboolean('samba', 'use-lmhash')



"""
Test cases for ldaptor.dns
"""

from twisted.trial import unittest
from ldaptor import dns

class NetmaskToNumbits(unittest.TestCase):
    def test_classA(self):
        self.assertEqual(dns.netmaskToNumbits('255.0.0.0'), 8)

    def test_classB(self):
        self.assertEqual(dns.netmaskToNumbits('255.255.0.0'), 16)

    def test_classC(self):
        self.assertEqual(dns.netmaskToNumbits('255.255.255.0'), 24)

    def test_host(self):
        self.assertEqual(dns.netmaskToNumbits('255.255.255.255'), 32)

    def test_numbits(self):
        for i in range(0, 33):
            self.assertEqual(dns.netmaskToNumbits(str(i)), i)

    def test_CIDR(self):
        for i in range(0, 33):
            mask = dns.ntoa(dns.aton(i))
            self.assertEqual(dns.netmaskToNumbits(mask), i)

class PtrSoaName(unittest.TestCase):
    def test_classA(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '255.0.0.0'),
                          '1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '8'),
                          '1.in-addr.arpa.')

    def test_classB(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '255.255.0.0'),
                          '2.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '16'),
                          '2.1.in-addr.arpa.')

    def test_classC(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '255.255.255.0'),
                          '3.2.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '24'),
                          '3.2.1.in-addr.arpa.')

    def test_CIDR_9(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '9'),
                          '0/9.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.200.3.4', '9'),
                          '128/9.1.in-addr.arpa.')

    def test_CIDR_12(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '12'),
                          '0/12.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.200.3.4', '12'),
                          '192/12.1.in-addr.arpa.')

    def test_CIDR_13(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '13'),
                          '0/13.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.200.3.4', '13'),
                          '200/13.1.in-addr.arpa.')

    def test_CIDR_15(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '15'),
                          '2/15.1.in-addr.arpa.')
        self.assertEqual(dns.ptrSoaName('1.200.3.4', '15'),
                          '200/15.1.in-addr.arpa.')

    def test_CIDR_29(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '29'),
                          '0/29.3.2.1.in-addr.arpa.')

    def test_CIDR_30(self):
        self.assertEqual(dns.ptrSoaName('1.2.3.4', '30'),
                          '4/30.3.2.1.in-addr.arpa.')



from io import BytesIO

from twisted.python import failure
from twisted.internet import reactor, protocol
from twisted.test import testutils


class FakeTransport(protocol.FileWrapper):
    disconnecting = False
    disconnect_done = False

    def __init__(self):
        self.data = BytesIO()
        protocol.FileWrapper.__init__(self, self.data)


class FasterIOPump(testutils.IOPump):
    def pump(self):
        """Move data back and forth.

        Returns whether any data was moved.
        """
        self.clientIO.seek(0)
        self.serverIO.seek(0)
        cData = self.clientIO.read()
        sData = self.serverIO.read()
        self.clientIO.seek(0)
        self.serverIO.seek(0)
        self.clientIO.truncate()
        self.serverIO.truncate()
        self.server.dataReceived(cData)
        self.client.dataReceived(sData)
        if cData or sData:
            return 1
        else:
            return 0


class IOPump(FasterIOPump):
    active = []

    def __init__(self,
                 client, server,
                 clientTransport, serverTransport):
        self.clientTransport = clientTransport
        self.serverTransport = serverTransport
        testutils.IOPump.__init__(self,
                                  client=client,
                                  server=server,
                                  clientIO=clientTransport.data,
                                  serverIO=serverTransport.data)
        self.active.append(self)


def returnConnected(server, client):
    """Take two Protocol instances and connect them.
    """
    clientTransport = FakeTransport()
    client.makeConnection(clientTransport)
    serverTransport = FakeTransport()
    server.makeConnection(serverTransport)
    pump = IOPump(client, server,
                  clientTransport,
                  serverTransport)
    # Challenge-response authentication:
    pump.flush()
    # Uh...
    pump.flush()
    return pump


def _getDeferredResult(d):
    resultSet = []
    d.addBoth(resultSet.append)
    while not resultSet:
        for pump in IOPump.active:
            pump.pump()
        reactor.iterate()
    return resultSet[0]


def pumpingDeferredResult(d):
    result = _getDeferredResult(d)
    return (
        result.raiseException()
        if isinstance(result, failure.Failure)
        else result
    )



"""DNS-related utilities."""

from socket import inet_aton, inet_ntoa
import struct


def aton_octets(ip):
    s = inet_aton(ip)
    return struct.unpack('!I', s)[0]


def aton_numbits(num):
    n = 0
    while num>0:
        n>>=1
        n |= 2**31
        num-=1
    return n


def aton(ip):
    try:
        i=int(ip)
    except ValueError:
        return aton_octets(ip)
    else:
        return aton_numbits(i)


def ntoa(n):
    s = struct.pack('!I', n)
    ip = inet_ntoa(s)
    return ip


def netmaskToNumbits(netmask):
    bits = aton(netmask)
    i = 2**31
    n = 0
    while bits and i > 0:
        if (bits & i) == 0:
            if bits:
                raise RuntimeError("Invalid netmask: %s" % netmask)
        n += 1
        bits -= i
        i = i >> 1
    return n

def ptrSoaName(ip, netmask):
    """
    Convert an IP address and netmask to a CIDR delegation
    -style zone name.
    """
    net = aton(ip) & aton(netmask)

    nmBits = netmaskToNumbits(netmask)
    bytes, bits = divmod(nmBits, 8)
    octets = ntoa(net).split('.')
    octets.reverse()
    if not bits:
        octets = octets[-bytes:]
    else:
        partial = octets[-bytes-1]
        octets = octets[-bytes:]
        octets.insert(0, '%s/%d' % (partial, nmBits))

    return '.'.join(octets)+'.in-addr.arpa.'



"""
Test cases for ldaptor.protocols.ldap.distinguishedname module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import distinguishedname as dn


class TestCaseWithKnownValues(unittest.TestCase):
    knownValues = ()

    def testKnownValues(self):
        for s, l in self.knownValues:
            fromString = dn.DistinguishedName(s)
            listOfRDNs = []
            for av in l:
                listOfAttributeTypesAndValues = []
                for a,v in av:
                    listOfAttributeTypesAndValues.append(dn.LDAPAttributeTypeAndValue(attributeType=a, value=v))
                r=dn.RelativeDistinguishedName(listOfAttributeTypesAndValues)
                listOfRDNs.append(r)
            fromList = dn.DistinguishedName(listOfRDNs)

            self.assertEqual(fromString, fromList)

            fromStringToText = fromString.getText()
            fromListToText = fromList.getText()

            assert fromStringToText == fromListToText

            canon = fromStringToText
            # DNs equal their byte string representation. Note this does
            # not mean they equal all the possible string
            # representations -- just the canonical one.
            self.assertEqual(fromString, canon)
            self.assertEqual(fromList, canon)
            self.assertEqual(canon, fromString)
            self.assertEqual(canon, fromList)

            # DNs can be used interchangeably with their canonical
            # string representation as hash keys.
            self.assertEqual(hash(fromString), hash(canon))
            self.assertEqual(hash(fromList), hash(canon))
            self.assertEqual(hash(canon), hash(fromString))
            self.assertEqual(hash(canon), hash(fromList))


class LDAPDistinguishedName_Escaping(TestCaseWithKnownValues):
    knownValues = (

        ('', []),

        ('cn=foo', [[('cn', 'foo')]]),

        (r'cn=\,bar', [[('cn', r',bar')]]),
        (r'cn=foo\,bar', [[('cn', r'foo,bar')]]),
        (r'cn=foo\,', [[('cn', r'foo,')]]),

        (r'cn=\+bar', [[('cn', r'+bar')]]),
        (r'cn=foo\+bar', [[('cn', r'foo+bar')]]),
        (r'cn=foo\+', [[('cn', r'foo+')]]),

        (r'cn=\"bar', [[('cn', r'"bar')]]),
        (r'cn=foo\"bar', [[('cn', r'foo"bar')]]),
        (r'cn=foo\"', [[('cn', r'foo"')]]),

        (r'cn=\\bar', [[('cn', r'\bar')]]),
        (r'cn=foo\\bar', [[('cn', r'foo\bar')]]),
        (r'cn=foo\\', [[('cn', 'foo\\')]]),

        (r'cn=\<bar', [[('cn', r'<bar')]]),
        (r'cn=foo\<bar', [[('cn', r'foo<bar')]]),
        (r'cn=foo\<', [[('cn', r'foo<')]]),

        (r'cn=\>bar', [[('cn', r'>bar')]]),
        (r'cn=foo\>bar', [[('cn', r'foo>bar')]]),
        (r'cn=foo\>', [[('cn', r'foo>')]]),

        (r'cn=\;bar', [[('cn', r';bar')]]),
        (r'cn=foo\;bar', [[('cn', r'foo;bar')]]),
        (r'cn=foo\;', [[('cn', r'foo;')]]),

        (r'cn=\#bar', [[('cn', r'#bar')]]),

        (r'cn=\ bar', [[('cn', r' bar')]]),

        (r'cn=bar\ ', [[('cn', r'bar ')]]),

        (r'cn=test+owner=uid\=foo\,ou\=depar'
         +r'tment\,dc\=example\,dc\=com,dc=ex'
         +r'ample,dc=com', [[('cn', r'test'),
                             ('owner', r'uid=foo,ou=depart'
                              +r'ment,dc=example,dc=com'),
                             ],
                            [('dc', r'example')],
                            [('dc', r'com')]]),

        (r'cn=bar,dc=example,dc=com', [[('cn', 'bar')],
                                       [('dc', 'example')],
                                       [('dc', 'com')]]),
        (r'cn=bar, dc=example, dc=com', [[('cn', 'bar')],
                                         [('dc', 'example')],
                                         [('dc', 'com')]]),
        (r'cn=bar,  dc=example,dc=com', [[('cn', 'bar')],
                                         [('dc', 'example')],
                                         [('dc', 'com')]]),

        )

    def testOpenLDAPEqualsEscape(self):
        """Slapd wants = to be escaped in RDN attributeValues."""
        got = dn.DistinguishedName(listOfRDNs=[
            dn.RelativeDistinguishedName(
            attributeTypesAndValues=[
            dn.LDAPAttributeTypeAndValue(attributeType='cn', value=r'test'),
            dn.LDAPAttributeTypeAndValue(attributeType='owner', value=r'uid=foo,ou=depart'
                                         +r'ment,dc=example,dc=com'),
                                     ]),

            dn.RelativeDistinguishedName('dc=example'),
            dn.RelativeDistinguishedName('dc=com'),
            ])
        got = got.getText()
        self.assertEqual(got,
                          'cn=test+owner=uid\=foo\,ou\=depar'
                          +'tment\,dc\=example\,dc\=com,dc=ex'
                          +'ample,dc=com')


class LDAPDistinguishedName_RFC2253_ExamplesBytes(TestCaseWithKnownValues):
    """
    It can be initialized from text/Unicode input as long as they contain
    ASCII only characters.
    """
    knownValues = (

        ('CN=Steve Kille,O=Isode Limited,C=GB',
         [[('CN', 'Steve Kille')],
          [('O', 'Isode Limited')],
          [('C', 'GB')]]),


        ('OU=Sales+CN=J. Smith,O=Widget Inc.,C=US',
         [[('OU', 'Sales'),
           ('CN', 'J. Smith')],
          [('O', 'Widget Inc.')],
          [('C', 'US')]]),

        (r'CN=L. Eagle,O=Sue\, Grabbit and Runn,C=GB',
         [[('CN', 'L. Eagle')],
          [('O', 'Sue, Grabbit and Runn')],
          [('C', 'GB')]]),

        (r'CN=Before\0DAfter,O=Test,C=GB',
         [[('CN', 'Before\x0dAfter')],
          [('O', 'Test')],
          [('C', 'GB')]]),

        (r'1.3.6.1.4.1.1466.0=#04024869,O=Test,C=GB',
         [[('1.3.6.1.4.1.1466.0', '#04024869')],
          [('O', 'Test')],
          [('C', 'GB')]]),
        )


class LDAPDistinguishedName_UTF8_Init(TestCaseWithKnownValues):
    """
    It can be initialized from an UTF-8 encoded data and it will
    keep the representation as UTF-8.
    """
    knownValues = (
        (u'SN=Lu\u010di\u0107'.encode('utf-8'),
         [[(b'SN', u'Lu\u010di\u0107'.encode('utf-8'))]]),
        )



class LDAPDistinguishedName_InitialSpaces(TestCaseWithKnownValues):
    """
    The spaces which are not escapes are stripped.
    """
    knownValues = (

        ('cn=foo, ou=bar,  dc=quux, \ attributeThatStartsWithSpace=Value',
         [[('cn', 'foo')],
          [('ou', 'bar')],
          [('dc', 'quux')],
          [(' attributeThatStartsWithSpace', 'Value')]]),

        )

class LDAPDistinguishedName_DomainName(unittest.TestCase):
    def testNonDc(self):
        d=dn.DistinguishedName('cn=foo,o=bar,c=us')
        assert d.getDomainName() is None

    def testNonTrailingDc(self):
        d=dn.DistinguishedName('cn=foo,o=bar,dc=foo,c=us')
        assert d.getDomainName() is None

    def testSimple_ExampleCom(self):
        d=dn.DistinguishedName('dc=example,dc=com')
        assert d.getDomainName() == 'example.com'

    def testSimple_SubExampleCom(self):
        d=dn.DistinguishedName('dc=sub,dc=example,dc=com')
        assert d.getDomainName() == 'sub.example.com'

    def testSimple_HostSubExampleCom(self):
        d=dn.DistinguishedName('cn=host,dc=sub,dc=example,dc=com')
        assert d.getDomainName() == 'sub.example.com'

    def testInterleaved_SubHostSubExampleCom(self):
        d=dn.DistinguishedName('dc=sub2,cn=host,dc=sub,dc=example,dc=com')
        assert d.getDomainName() == 'sub.example.com'

class LDAPDistinguishedName_contains(unittest.TestCase):
    shsec=dn.DistinguishedName('dc=sub2,cn=host,dc=sub,dc=example,dc=com')
    hsec=dn.DistinguishedName('cn=host,dc=sub,dc=example,dc=com')
    sec=dn.DistinguishedName('dc=sub,dc=example,dc=com')
    ec=dn.DistinguishedName('dc=example,dc=com')
    c=dn.DistinguishedName('dc=com')

    soc=dn.DistinguishedName('dc=sub,dc=other,dc=com')
    oc=dn.DistinguishedName('dc=other,dc=com')

    other=dn.DistinguishedName('o=foo,c=US')

    root=dn.DistinguishedName('')

    def test_selfContainment(self):
        assert self.c.contains(self.c)
        assert self.ec.contains(self.ec)
        assert self.sec.contains(self.sec)
        assert self.hsec.contains(self.hsec)
        assert self.shsec.contains(self.shsec)

        assert self.soc.contains(self.soc)
        assert self.oc.contains(self.oc)

        assert self.root.contains(self.root)

        assert self.other.contains(self.other)

    def test_realContainment(self):
        assert self.c.contains(self.ec)
        assert self.c.contains(self.sec)
        assert self.c.contains(self.hsec)
        assert self.c.contains(self.shsec)

        assert self.ec.contains(self.sec)
        assert self.ec.contains(self.hsec)
        assert self.ec.contains(self.shsec)

        assert self.sec.contains(self.hsec)
        assert self.sec.contains(self.shsec)

        assert self.hsec.contains(self.shsec)

        assert self.c.contains(self.oc)
        assert self.c.contains(self.soc)
        assert self.oc.contains(self.soc)

        for x in (self.shsec, self.hsec, self.sec, self.ec, self.c,
                  self.soc, self.oc, self.other):
            assert self.root.contains(x)

    def test_nonContainment_parents(self):
        assert not self.shsec.contains(self.hsec)
        assert not self.shsec.contains(self.sec)
        assert not self.shsec.contains(self.ec)
        assert not self.shsec.contains(self.c)

        assert not self.hsec.contains(self.sec)
        assert not self.hsec.contains(self.ec)
        assert not self.hsec.contains(self.c)

        assert not self.sec.contains(self.ec)
        assert not self.sec.contains(self.c)

        assert not self.ec.contains(self.c)
        assert not self.soc.contains(self.oc)

        for x in (self.shsec, self.hsec, self.sec, self.ec, self.c,
                  self.soc, self.oc, self.other):
            assert not x.contains(self.root)

    def test_nonContainment_nonParents(self):
        groups=([self.shsec, self.hsec, self.sec, self.ec],
                [self.soc, self.oc],
                [self.other])
        for g1 in groups:
            for g2 in groups:
                if g1!=g2:
                    for i1 in g1:
                        for i2 in g2:
                            assert not i1.contains(i2)
        assert not self.c.contains(self.other)
        assert not self.other.contains(self.c)

class LDAPDistinguishedName_Malformed(unittest.TestCase):
    def testMalformed(self):
        self.assertRaises(dn.InvalidRelativeDistinguishedName,
                          dn.DistinguishedName,
                          'foo')
        self.assertRaises(dn.InvalidRelativeDistinguishedName,
                          dn.DistinguishedName,
                          'foo,dc=com')
        self.assertRaises(dn.InvalidRelativeDistinguishedName,
                          dn.DistinguishedName,
                          'ou=something,foo')
        self.assertRaises(dn.InvalidRelativeDistinguishedName,
                          dn.DistinguishedName,
                          'foo,foo')

class LDAPDistinguishedName_Prettify(unittest.TestCase):
    def testPrettifySpaces(self):
        """DistinguishedName(...).getText() prettifies the DN by removing extra whitespace."""
        d=dn.DistinguishedName('cn=foo, o=bar,  c=us')
        assert d.getText() == u'cn=foo,o=bar,c=us'

class DistinguishedName_Init(unittest.TestCase):
    def testGetText(self):
        d=dn.DistinguishedName('dc=example,dc=com')
        self.assertEqual(d.getText(), u'dc=example,dc=com')

    def testDN(self):
        proto=dn.DistinguishedName('dc=example,dc=com')
        d=dn.DistinguishedName(proto)
        self.assertEqual(d.getText(), u'dc=example,dc=com')

    def testEqualToByteString(self):
        """
        DistinguishedName is equal to its bytes representation
        """
        d = dn.DistinguishedName('dc=example,dc=com')
        self.assertEqual(d, b'dc=example,dc=com')

    def testEqualToString(self):
        """
        DistinguishedName is equal to its unicode representation
        """
        d = dn.DistinguishedName('dc=example,dc=com')
        self.assertEqual(d, u'dc=example,dc=com')


class RelativeDistinguishedName_Init(unittest.TestCase):
    def testGetText(self):
        rdn=dn.RelativeDistinguishedName('dc=example')
        self.assertEqual(rdn.getText(), u'dc=example')

    def testRDN(self):
        proto=dn.RelativeDistinguishedName('dc=example')
        rdn=dn.RelativeDistinguishedName(proto)
        self.assertEqual(rdn.getText(), u'dc=example')

class DistinguishedName_Comparison(unittest.TestCase):
    """
    Tests for comparing DistinguishedName.
    """

    def test_parent_child(self):
        """
        The parent is greater than the child.
        """
        dn1=dn.DistinguishedName('dc=example,dc=com')
        dn2=dn.DistinguishedName('dc=and,dc=example,dc=com')

        self.assertLess(dn2, dn1)
        self.assertGreater(dn1, dn2)



"""
Test cases for ldaptor.protocols.ldap.ldapserver module.
"""

from twisted.trial import unittest
from ldaptor import inmemory
from ldaptor.protocols import pureldap, pureber
from ldaptor.protocols.ldap import ldapsyntax

class TestEntryMatch(unittest.TestCase):
    def test_matchAll(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilterMatchAll)
        self.assertEqual(result, True)

    def test_present_match(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_present('aValue'))
        self.assertEqual(result, True)


    def test_present_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_present('noSuchValue'))
        self.assertEqual(result, False)

    def test_and_match(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(
            pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_present('aValue'),
            pureldap.LDAPFilter_present('bValue'),
            ]))
        self.assertEqual(result, True)

    def test_and_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(
            pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_present('cValue'),
            pureldap.LDAPFilter_present('dValue'),
            ]))
        self.assertEqual(result, False)

    def test_or_match(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(
            pureldap.LDAPFilter_or([
            pureldap.LDAPFilter_present('cValue'),
            pureldap.LDAPFilter_present('bValue'),
            ]))
        self.assertEqual(result, True)

    def test_or_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(
            pureldap.LDAPFilter_or([
            pureldap.LDAPFilter_present('cValue'),
            pureldap.LDAPFilter_present('dValue'),
            ]))
        self.assertEqual(result, False)

    def test_not(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(
            pureldap.LDAPFilter_not(
            pureldap.LDAPFilter_or([
            pureldap.LDAPFilter_present('cValue'),
            pureldap.LDAPFilter_present('dValue'),
            ])))
        self.assertEqual(result, True)

    def test_equality_match(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureber.BEROctetString('aValue'),
            assertionValue=pureber.BEROctetString('a')))
        self.assertEqual(result, True)

    def test_equality_match_caseInsensitive(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureber.BEROctetString('avaLUe'),
            assertionValue=pureber.BEROctetString('A')))
        self.assertEqual(result, True)


    def test_equality_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureber.BEROctetString('aValue'),
            assertionValue=pureber.BEROctetString('b')))
        self.assertEqual(result, False)

    def test_substrings_match(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match2(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['abcde'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match3(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['abcde'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            pureldap.LDAPFilter_substrings_any('c'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match4(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['abcde'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            pureldap.LDAPFilter_substrings_any('b'),
            pureldap.LDAPFilter_substrings_any('c'),
            pureldap.LDAPFilter_substrings_any('d'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match5(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['aoeuboeucoeudoeue'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            pureldap.LDAPFilter_substrings_any('b'),
            pureldap.LDAPFilter_substrings_any('c'),
            pureldap.LDAPFilter_substrings_any('d'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match6(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['aBCdE'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('A'),
            pureldap.LDAPFilter_substrings_any('b'),
            pureldap.LDAPFilter_substrings_any('C'),
            pureldap.LDAPFilter_substrings_any('D'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_match7(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['Foo'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('f'),
            ]))
        self.assertEqual(result, True)

    def test_substrings_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['a'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('bad'),
            pureldap.LDAPFilter_substrings_any('dog'),
            pureldap.LDAPFilter_substrings_any('no'),
            pureldap.LDAPFilter_substrings_final('bone'),
            ]))
        self.assertEqual(result, False)

    def test_substrings_noMatch2(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['aoeuboeucoeudoeue'],
            'bValue': ['b'],
            })
        result = o.match(pureldap.LDAPFilter_substrings(
            type='aValue',
            substrings=[
            pureldap.LDAPFilter_substrings_initial('a'),
            pureldap.LDAPFilter_substrings_any('b'),
            pureldap.LDAPFilter_substrings_any('Z'),
            pureldap.LDAPFilter_substrings_any('d'),
            pureldap.LDAPFilter_substrings_final('e'),
            ]))
        self.assertEqual(result, False)

    def test_greaterOrEqual_noMatch_nosuchattr(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_greaterOrEqual(pureber.BEROctetString('foo'),
                                                            pureber.BERInteger(42)))
        self.assertEqual(result, False)

    def test_greaterOrEqual_match_greater(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_greaterOrEqual(pureber.BEROctetString('num'),
                                                            pureber.BERInteger(3)))
        self.assertEqual(result, True)

    def test_greaterOrEqual_match_equal(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_greaterOrEqual(pureber.BEROctetString('num'),
                                                            pureber.BERInteger(4)))
        self.assertEqual(result, True)

    def test_greaterOrEqual_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'bValue': [4],
            })
        result = o.match(pureldap.LDAPFilter_greaterOrEqual(pureber.BEROctetString('num'),
                                                            pureber.BERInteger(5)))
        self.assertEqual(result, False)


    def test_lessOrEqual_noMatch_nosuchattr(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_lessOrEqual(pureber.BEROctetString('foo'),
                                                         pureber.BERInteger(42)))
        self.assertEqual(result, False)

    def test_lessOrEqual_match_less(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_lessOrEqual(pureber.BEROctetString('num'),
                                                         pureber.BERInteger(5)))
        self.assertEqual(result, True)

    def test_lessOrEqual_match_equal(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_lessOrEqual(pureber.BEROctetString('num'),
                                                         pureber.BERInteger(4)))
        self.assertEqual(result, True)

    def test_lessOrEqual_noMatch(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        result = o.match(pureldap.LDAPFilter_lessOrEqual(pureber.BEROctetString('num'),
                                                         pureber.BERInteger(3)))
        self.assertEqual(result, False)

    def test_extensibleMatch4(self):
        """
        An extensibleMatch filter that uses DN attributes matches an entry
        based on its OU.
        See RFC4511 section 4.5.1.
        """
        m = pureldap.LDAPFilter_extensibleMatch(
            matchingRule=None, 
            type=pureldap.LDAPMatchingRuleAssertion_type(value='ou'), 
            matchValue=pureldap.LDAPMatchingRuleAssertion_matchValue(value='fings'), 
            dnAttributes=pureldap.LDAPMatchingRuleAssertion_dnAttributes(value=255))
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn = 'cn=foo,ou=fings,dc=example,dc=com',
            attributes={
                'objectClass': ['a', 'b'],
                'aValue': ['b'],
                'num': [4],
            })
        result = o.match(m)
        self.assertEqual(result, True)

    def test_extensibleMatch4_noMatch(self):
        """
        An extensibleMatch filter that uses DN attributes does not match an entry
        based on its OU.
        See RFC4511 section 4.5.1.
        """
        m = pureldap.LDAPFilter_extensibleMatch(
            matchingRule=None, 
            type=pureldap.LDAPMatchingRuleAssertion_type(value='ou'), 
            matchValue=pureldap.LDAPMatchingRuleAssertion_matchValue(value='fings'), 
            dnAttributes=pureldap.LDAPMatchingRuleAssertion_dnAttributes(value=255))
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn = 'cn=foo,ou=uvvers,dc=example,dc=com',
            attributes={
                'objectClass': ['a', 'b'],
                'aValue': ['b'],
                'num': [4],
            })
        result = o.match(m)
        self.assertEqual(result, False)

    def test_notImplemented(self):
        o=inmemory.ReadOnlyInMemoryLDAPEntry(dn='cn=foo,dc=example,dc=com',
                                             attributes={
            'objectClass': ['a', 'b'],
            'aValue': ['b'],
            'num': [4],
            })
        class UnknownMatch(object): pass
        unknownMatch = UnknownMatch()
        self.assertRaises(ldapsyntax.MatchNotImplemented,
                          o.match, unknownMatch)

# TODO LDAPFilter_approxMatch
# TODO LDAPFilter_extensibleMatch



import sys

from twisted.internet import defer
from twisted.internet.endpoints import clientFromString, connectProtocol
from twisted.internet.task import react
from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry


@defer.inlineCallbacks
def onConnect(client):
    # The following arguments may be also specified as unicode strings
    # but it is recommended to use byte strings for ldaptor objects
    basedn = b'dc=example,dc=org'
    binddn = b'cn=bob,ou=people,dc=example,dc=org'
    bindpw = b'secret'
    query = b'(cn=bob)'
    try:
        yield client.bind(binddn, bindpw)
    except Exception as ex:
        print(ex)
        raise
    o = LDAPEntry(client, basedn)
    results = yield o.search(filterText=query)
    for entry in results:
        print(entry.getLDIF())


def onError(err):
    err.printDetailedTraceback(file=sys.stderr)


def main(reactor):
    endpoint_str = "tcp:host=127.0.0.1:port=8080"
    e = clientFromString(reactor, endpoint_str)
    d = connectProtocol(e, LDAPClient())
    d.addCallback(onConnect)
    d.addErrback(onError)
    return d


react(main)



"""
Test cases for ldaptor.protocols.ldap.ldifdelta
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldifdelta
from ldaptor import delta, entry

class LDIFDeltaDriver(ldifdelta.LDIFDelta):
    def __init__(self):
        self.listOfCompleted = []
    def gotEntry(self, obj):
        self.listOfCompleted.append(obj)


"""
changerecord             = "changetype:" FILL
                           (change-add / change-delete /
                            change-modify / change-moddn)

change-add               = "add"                SEP 1*attrval-spec

change-delete            = "delete"             SEP

change-moddn             = ("modrdn" / "moddn") SEP
                            "newrdn:" (    FILL rdn /
                                       ":" FILL base64-rdn) SEP
                            "deleteoldrdn:" FILL ("0" / "1")  SEP
                            0*1("newsuperior:"
                            (    FILL distinguishedName /
                             ":" FILL base64-distinguishedName) SEP)

change-modify            = "modify"             SEP *mod-spec

mod-spec                 = ("add:" / "delete:" / "replace:")
                           FILL AttributeDescription SEP
                           *attrval-spec
                           "-" SEP
"""


"""
version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete

"""

"""
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modrdn  #OR moddn
newrdn: rdn
deleteoldrdn: 0 #OR 1
#0..1 newsuperior: distinguishedName

"""


class TestLDIFDeltaParsing(unittest.TestCase):
    def testModification_empty(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify

""")
        proto.connectionLost()
        self.assertEqual(proto.listOfCompleted,
                         [
            delta.ModifyOp(dn='cn=foo,dc=example,dc=com'),
            ])

    def testModification_oneAdd(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar
-

""")
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [delta.ModifyOp(dn=b'cn=foo,dc=example,dc=com',
                            modifications=[delta.Add(b'foo', [b'bar']),
                                           ]),
             ])

    def testModification_twoAdds(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar
-
add: thud
thud: quux
thud: baz
-

""")
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [delta.ModifyOp(dn=b'cn=foo,dc=example,dc=com',
                            modifications=[delta.Add(b'foo', [b'bar']),
                                           delta.Add(b'thud', [b'quux', b'baz']),
                                           ]),
             ])

    def testModification_complex(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
delete: foo
foo: bar
-
delete: garply
-
add: thud
thud: quux
thud: baz
-
replace: waldo
-
add: foo
foo: baz
-
replace: thud
thud: xyzzy
-
add: silly
-

""")
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [delta.ModifyOp(dn=b'cn=foo,dc=example,dc=com',
                            modifications=[delta.Delete(b'foo', [b'bar']),
                                           delta.Delete(b'garply'),
                                           delta.Add(b'thud', [b'quux', b'baz']),
                                           delta.Replace(b'waldo'),
                                           delta.Add(b'foo', [b'baz']),
                                           delta.Replace(b'thud', [b'xyzzy']),
                                           delta.Add(b'silly'),
                                           ]),
             ])

    def testModification_fail_noDash_1(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(ldifdelta.LDIFDeltaModificationMissingEndDashError,
                          proto.dataReceived,
                          b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar

""")

    def testModification_fail_noDash_2(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(ldifdelta.LDIFDeltaModificationMissingEndDashError,
                          proto.dataReceived,
                          b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo

""")

    def testModification_fail_differentKey(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(ldifdelta.LDIFDeltaModificationDifferentAttributeTypeError,
                          proto.dataReceived,
                          b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
bar: quux
-

""")

    def testModification_fail_unknownModSpec(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(ldifdelta.LDIFDeltaUnknownModificationError,
                          proto.dataReceived,
                          b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
fiddle: foo
foo: bar
-

""")

    def testNoChangeType(self):
        """
        Raises an error is the changetype is not present.
        """
        proto = LDIFDeltaDriver()
        error = self.assertRaises(
            ldifdelta.LDIFDeltaMissingChangeTypeError,
            proto.dataReceived,
            b"""version: 1
dn: cn=foo,dc=example,dc=com
add: foo
foo: bar
-

""")
        self.assertEqual(
            (b'cn=foo,dc=example,dc=com', b'add', b'foo'), error.args)


    def testNoChangetTypeEmpty(self):
        """
        Raises an error is the changetype is not present even if no
        other command are specified for the DN.
        """
        proto = LDIFDeltaDriver()

        error = self.assertRaises(
            ldifdelta.LDIFDeltaMissingChangeTypeError,
            proto.dataReceived,
            b"""version: 1
dn: cn=foo,dc=example,dc=com

""")

        self.assertEqual((b'cn=foo,dc=example,dc=com',), error.args)


    def testAdd(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: add
foo: bar
thud: quux
thud: baz

""")
        proto.connectionLost()
        self.assertEqual(proto.listOfCompleted,
                         [delta.AddOp(entry.BaseLDAPEntry(
            dn=b'cn=foo,dc=example,dc=com',
            attributes={
            b'foo': [b'bar'],
            b'thud': [b'quux', b'baz'],
            }))])

    def testAdd_fail_noAttrvals(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(ldifdelta.LDIFDeltaAddMissingAttributesError,
                          proto.dataReceived, b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: add

""")


    def testDelete(self):
        """"
        Triggers a DeleteOp when the diff action is `delete`.
        """
        proto = LDIFDeltaDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete

""")
        proto.connectionLost()
        self.assertEqual(proto.listOfCompleted,
                         [delta.DeleteOp(dn=b'cn=foo,dc=example,dc=com')])


    def testDeleteMalformat(self):
        """"
        Raises an error when delete change has data after the changetype.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(
            ldifdelta.LDIFDeltaDeleteHasJunkAfterChangeTypeError
                ):
            proto.dataReceived(
b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete
foo: bar

""")


    def testMODRDN(self):
        """
        modrdn is not yet supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(NotImplementedError):
            proto.dataReceived(
b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: modrdn

""")


    def testMODDN(self):
        """
        moddn is not yet supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(NotImplementedError):
            proto.dataReceived(
b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: moddn

""")


    def testUnknownChnagetType(self):
        """
        Raises an error is the changetype is not supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(ldifdelta.LDIFDeltaUnknownChangeTypeError):
            proto.dataReceived(
b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: some-random-type

""")




from ldaptor.protocols.ldap.autofill import ObjectMissingObjectClassException

class Autofill_samba: #TODO baseclass
    def __init__(self, domainSID, fixedPrimaryGroupSID=None):
        self.domainSID = domainSID
        self.fixedPrimaryGroupSID = fixedPrimaryGroupSID

    def start(self, ldapObject):
        assert 'objectClass' in ldapObject
        if 'sambaSamAccount' not in ldapObject['objectClass']:
            raise ObjectMissingObjectClassException(ldapObject)

        assert 'sambaAcctFlags' not in ldapObject
        ldapObject['sambaAcctFlags'] = ['[UX         ]']
        assert 'sambaPwdLastSet' not in ldapObject
        ldapObject['sambaPwdLastSet'] = ['1']
        assert 'sambaLogonTime' not in ldapObject
        ldapObject['sambaLogonTime'] = ['0']
        assert 'sambaLogoffTime' not in ldapObject
        ldapObject['sambaLogoffTime'] = ['0']
        assert 'sambaPwdCanChange' not in ldapObject
        ldapObject['sambaPwdCanChange'] = ['0']
        assert 'sambaPwdMustChange' not in ldapObject
        ldapObject['sambaPwdMustChange'] = ['0']

        if self.fixedPrimaryGroupSID is not None:
            assert 'sambaPrimaryGroupSID' not in ldapObject
            ldapObject['sambaPrimaryGroupSID'] = ['%s-%d' % (
                self.domainSID, self.fixedPrimaryGroupSID)]

        # Handle attributeTypes that were added before we got
        # started. We know we don't defer in notify, so we can do a
        # simple loop here.
        for attributeType in ldapObject.keys():
            self.notify(ldapObject, attributeType)

    def notify(self, ldapObject, attributeType):
        # sambaSID=2*uidNumber+1000
        if attributeType == 'uidNumber':
            assert 'uidNumber' in ldapObject
            assert len(ldapObject['uidNumber']) == 1
            for uidNumber in ldapObject['uidNumber']:
                uidNumber = int(uidNumber)
                sid = '%s-%d' % (self.domainSID, uidNumber*2+1000)
                ldapObject['sambaSID'] = [str(sid)]
                return

        # sambaPrimaryGroupSID = fixed or 2*gidNumber+1001
        if (self.fixedPrimaryGroupSID is None
            and attributeType == 'gidNumber'):
            assert 'gidNumber' in ldapObject
            assert len(ldapObject['gidNumber']) == 1
            for gidNumber in ldapObject['gidNumber']:
                gidNumber = int(gidNumber)
                sid = '%s-%d' % (self.domainSID, gidNumber*2+1001)
                ldapObject['sambaPrimaryGroupSID'] = [str(sid)]
                return



"""
Test cases for ldaptor.protocols.ldap.autofill.posixAccount module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldapsyntax, autofill
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap.autofill import posixAccount
from ldaptor.testutil import LDAPClientTestDriver

class LDAPAutoFill_Posix(unittest.TestCase):
    def testMustHaveObjectClass(self):
        """Test that Autofill_posix fails unless object is a posixAccount."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntryWithAutoFill(client=client,
                                           dn='cn=foo,dc=example,dc=com',
                                           attributes={
            'objectClass': ['something', 'other'],
            })
        autoFiller = posixAccount.Autofill_posix(baseDN='dc=example,dc=com')

        d = o.addAutofiller(autoFiller)

        failure = self.failureResultOf(d)
        self.assertIsInstance(
          failure.value, autofill.ObjectMissingObjectClassException)
        client.assertNothingSent()

    def testDefaultSetting(self):
        """Test that fields get their default values."""

        client = LDAPClientTestDriver(
            # uid==1000 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1000 -> taken
            [   pureldap.LDAPSearchResultEntry(objectName='',
                                           attributes=[('objectClass',
                                                        ('foo',
                                                         'posixAccount',
                                                         'bar'))]),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],
            # gid==1500 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1250 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1125 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1062 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],
            # gid==1031 -> free
            [   pureldap.LDAPSearchResultEntry(objectName='',
                                               attributes=[('objectClass',
                                                            ('foo',
                                                             'posixAccount',
                                                             'bar'))]),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1046 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1038 -> taken
            [   pureldap.LDAPSearchResultEntry(objectName='',
                                               attributes=[('objectClass',
                                                            ('foo',
                                                             'posixAccount',
                                                             'bar'))]),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1042 -> free
            [   pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1040 -> taken
            [   pureldap.LDAPSearchResultEntry(objectName='',
                                               attributes=[('objectClass',
                                                            ('foo',
                                                             'posixAccount',
                                                             'bar'))]),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],

            # gid==1041 -> taken
            [   pureldap.LDAPSearchResultEntry(objectName='',
                                               attributes=[('objectClass',
                                                            ('foo',
                                                             'posixAccount',
                                                             'bar'))]),
                pureldap.LDAPSearchResultDone(resultCode=0,
                                              matchedDN='',
                                              errorMessage=''),],
           )

        o=ldapsyntax.LDAPEntryWithAutoFill(client=client,
                                           dn='cn=foo,dc=example,dc=com',
                                           attributes={
            'objectClass': ['posixAccount', 'other'],
            })

        d = o.addAutofiller(posixAccount.Autofill_posix(baseDN='dc=example,dc=com'))
        d.addCallback(self._cb_testDefaultSetting, client, o)
        return d

    def _cb_testDefaultSetting(self, val, client, o):
        client.assertSent(
            *[

            pureldap.LDAPSearchRequest(
            baseObject='dc=example,dc=com', scope=2,
            derefAliases=0, sizeLimit=1, timeLimit=0, typesOnly=0,
            filter=pureldap.LDAPFilter_equalityMatch(attributeDesc=pureldap.LDAPAttributeDescription(value='uidNumber'),
                                                     assertionValue=pureldap.LDAPAssertionValue(value='1000')),
            attributes=()),

            ] + [
            pureldap.LDAPSearchRequest(
            baseObject='dc=example,dc=com', scope=2,
            derefAliases=0, sizeLimit=1, timeLimit=0, typesOnly=0,
            filter=pureldap.LDAPFilter_equalityMatch(attributeDesc=pureldap.LDAPAttributeDescription(value='gidNumber'),
                                                     assertionValue=pureldap.LDAPAssertionValue(value=str(x))),
            attributes=())
            for x in (1000, 1500, 1250, 1125, 1062, 1031, 1046, 1038, 1042, 1040, 1041)])

        self.failUnless('loginShell' in o)
        self.failUnlessEqual(o['loginShell'], ['/bin/sh'])

        self.failUnless('uidNumber' in o)
        self.failUnlessEqual(o['uidNumber'], ['1000'])
        self.failUnless('gidNumber' in o)
        self.failUnlessEqual(o['gidNumber'], ['1042'])



import base64
import random

import six

from twisted.internet import defer
from twisted.python.util import InsensitiveDict
from zope.interface import implementer

from ldaptor import interfaces, attributeset, delta
from ldaptor._encoder import WireStrAlias, to_bytes, get_strings
from ldaptor.protocols.ldap import distinguishedname, ldif, ldaperrors


try:
    from hashlib import sha1
except ImportError:
    from sha import sha as sha1


def sshaDigest(passphrase, salt=None):
    """
    Return the salted SHA for `passphrase` which is passed as bytes.
    """
    if salt is None:
        salt = ''
        for i in range(8):
            salt += chr(random.randint(0, 127))
        salt = salt.encode('ascii')

    s = sha1()
    s.update(passphrase)
    s.update(salt)
    encoded = base64.encodestring(s.digest() + salt).rstrip()
    crypt = b'{SSHA}' + encoded
    return crypt


@implementer(interfaces.ILDAPEntry)
class BaseLDAPEntry(WireStrAlias):
    dn = None
    _object_class_keys = set(get_strings('objectClass'))
    _object_class_lower_keys = set(get_strings('objectclass'))
    _user_password_keys = set(get_strings('userPassword'))

    def __init__(self, dn, attributes={}):
        """

        Initialize the object.

        @param dn: Distinguished Name of the object, as a string.

        @param attributes: Attributes of the object. A dictionary of
        attribute types to list of attribute values.

        """
        self._attributes = InsensitiveDict()
        self.dn = distinguishedname.DistinguishedName(dn)

        for k, vs in attributes.items():
            if k not in self._attributes:
                self._attributes[k] = []
            self._attributes[k].extend(vs)

        for k, vs in self._attributes.items():
            self._attributes[k] = self.buildAttributeSet(k, vs)

    def buildAttributeSet(self, key, values):
        return attributeset.LDAPAttributeSet(key, values)

    def __getitem__(self, key):
        for k in get_strings(key):
            if k in self._attributes:
                return self._attributes[k]
        raise KeyError(key)

    def get(self, key, default=None):
        for k in get_strings(key):
            if k in self._attributes:
                return self._attributes[k]
        return default

    def has_key(self, key):
        for k in get_strings(key):
            if k in self._attributes:
                return True
        return False

    def __contains__(self, key):
        return self.has_key(key)

    def __iter__(self):
        for key in self._attributes.iterkeys():
            yield key

    def keys(self):
        a = []
        for key in self._object_class_keys:
            if key in self._attributes:
                a.append(key)
        l = list(self._attributes.keys())
        l.sort(key=to_bytes)
        for key in l:
            if key.lower() not in self._object_class_lower_keys:
                a.append(key)
        return a

    def items(self):
        a = []

        for key in self._object_class_keys:
            objectClasses = list(self._attributes.get(key, []))
            objectClasses.sort(key=to_bytes)
            if objectClasses:
                a.append((key, objectClasses))

        l = list(self._attributes.items())
        l.sort(key=lambda x: to_bytes(x[0]))
        for key, values in l:
            if key.lower() not in self._object_class_lower_keys:
                vs = list(values)
                vs.sort()
                a.append((key, vs))

        return a

    def toWire(self):
        a = []

        for key in self._object_class_keys:
            objectClasses = list(self._attributes.get(key, []))
            objectClasses.sort(key=to_bytes)
            a.append((key, objectClasses))

        items_gen = ((key, self[key]) for key in self)
        items = sorted(items_gen, key=lambda x: to_bytes(x[0]))
        for key, values in items:
            if key.lower() not in self._object_class_lower_keys:
                vs = list(values)
                vs.sort()
                a.append((key, vs))
        return ldif.asLDIF(self.dn.getText(), a)

    def getLDIF(self):
        return self.toWire().decode('utf-8')

    def __eq__(self, other):
        if not isinstance(other, BaseLDAPEntry):
            return NotImplemented
        if self.dn != other.dn:
            return 0

        my = sorted((key for key in self), key=to_bytes)
        its = sorted((key for key in other), key=to_bytes)
        if my != its:
            return 0
        for key in my:
            myAttr = self[key]
            itsAttr = other[key]
            if myAttr != itsAttr:
                return 0
        return 1

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return len(self.keys())

    def __bool__(self):
        return True

    def __nonzero__(self):
        return self.__bool__()

    def __repr__(self):
        keys = sorted((key for key in self), key=to_bytes)
        a = []
        for key in keys:
            a.append('%s: %s' % (repr(key), repr(list(self[key]))))
        attributes = ', '.join(a)
        dn = to_bytes(self.dn.getText()) if six.PY2 else self.dn.getText()
        return '%s(%s, {%s})' % (
            self.__class__.__name__,
            repr(dn),
            attributes)

    def diff(self, other):
        """
        Compute differences between this and another LDAP entry.

        @param other: An LDAPEntry to compare to.

        @return: None if equal, otherwise a ModifyOp that would make
        this entry look like other.
        """
        assert self.dn == other.dn
        if self == other:
            return None

        r = []

        myKeys = set(key for key in self)
        otherKeys = set(key for key in other)

        addedKeys = list(otherKeys - myKeys)
        addedKeys.sort(key=to_bytes)  # for reproducability only
        for added in addedKeys:
            r.append(delta.Add(added, other[added]))

        deletedKeys = list(myKeys - otherKeys)
        deletedKeys.sort(key=to_bytes)  # for reproducability only
        for deleted in deletedKeys:
            r.append(delta.Delete(deleted, self[deleted]))

        sharedKeys = list(myKeys & otherKeys)
        sharedKeys.sort(key=to_bytes)  # for reproducability only
        for shared in sharedKeys:

            addedValues = list(other[shared] - self[shared])
            if addedValues:
                addedValues.sort(key=to_bytes)  # for reproducability only
                r.append(delta.Add(shared, addedValues))

            deletedValues = list(self[shared] - other[shared])
            if deletedValues:
                deletedValues.sort(key=to_bytes)  # for reproducability only
                r.append(delta.Delete(shared, deletedValues))

        return delta.ModifyOp(dn=self.dn, modifications=r)

    def bind(self, password):
        return defer.maybeDeferred(self._bind, password)

    def _bind(self, password):
        password = to_bytes(password)
        for key in self._user_password_keys:
            for digest in self.get(key, ()):
                digest = to_bytes(digest)
                if digest.startswith(b'{SSHA}'):
                    raw = base64.decodestring(digest[len(b'{SSHA}'):])
                    salt = raw[20:]
                    got = sshaDigest(password, salt)
                    if got == digest:
                        return self
                else:
                    # Plaintext
                    if digest == password:
                        return self
        raise ldaperrors.LDAPInvalidCredentials()

    def hasMember(self, dn):
        for memberDN in self.get('member', []):
            if memberDN == dn:
                return True
        return False

    def __hash__(self):
        # FIXME:https://github.com/twisted/ldaptor/issues/101
        # The hash should take into consideration any attribute used to
        # decide the equality.
        return hash(self.dn)


@implementer(interfaces.IEditableLDAPEntry)
class EditableLDAPEntry(BaseLDAPEntry):

    def __setitem__(self, key, value):
        new = self.buildAttributeSet(key, value)
        self._attributes[key] = new

    def __delitem__(self, key):
        del self._attributes[key]

    def undo(self):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()

    def move(self, newDN):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def setPassword(self, newPasswd, salt=None):
        """
        Update the password for the entry with a new password and salt passed
        as bytes.
        """
        crypt = sshaDigest(newPasswd, salt)
        for key in self._user_password_keys:
            if key in self:
                self[key] = [crypt]
        else:
            self[b'userPassword'] = [crypt]



"""
Test cases for ldaptor.protocols.ldap.ldif module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldifprotocol, distinguishedname


class FixStringRepresentation(object):
    """
    A simple object which has a fix string representation.
    """
    def __str__(self):
        return "Here I am!"


class TestLDIFParseError(unittest.TestCase):
    """
    Unit tests for LDIFParseError which is hte base for all the other
    LDIF errors.
    """

    def testInitNoArgs(self):
        """
        It can be initialized without arguments and will have the
        docstring as the string representation.
        """
        sut = ldifprotocol.LDIFParseError()

        result = str(sut)

        self.assertEqual('Error parsing LDIF.', result)

    def testInitWithArgs(self):
        """
        When initialized with arguments it will use the docstring as
        base and include all the arguments.
        """
        sut = ldifprotocol.LDIFParseError(
            1, 'test', True, FixStringRepresentation())

        result = str(sut)

        self.assertEqual(
            'Error parsing LDIF: 1: test: True: Here I am!.', result)


class LDIFDriver(ldifprotocol.LDIF):
    def __init__(self):
        self.listOfCompleted = []
    def gotEntry(self, obj):
        self.listOfCompleted.append(obj)

class TestLDIFParsing(unittest.TestCase):
    def testFromLDIF(self):
        proto = LDIFDriver()
        for line in (

            "dn: cn=foo,dc=example,dc=com",
            "objectClass: a",
            "objectClass: b",
            "aValue: a",
            "aValue: b",
            "bValue: c",
            "",

            "dn: cn=bar,dc=example,dc=com",
            "objectClass: c",
            "aValue:: IEZPTyE=",
            "aValue: b",
            "bValue: C",
            "",

            ):
            proto.lineReceived(line.encode('ascii'))

        self.failUnlessEqual(len(proto.listOfCompleted), 2)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'a', b'b'])
        self.failUnlessEqual(o[b'aValue'], [b'a', b'b'])
        self.failUnlessEqual(o[b'bValue'], [b'c'])

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=bar,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'c'])
        self.failUnlessEqual(o[b'aValue'], [b' FOO!', b'b'])
        self.failUnlessEqual(o[b'bValue'], [b'C'])

        self.failUnlessEqual(proto.listOfCompleted, [])

    def testSplitLines(self):
        """
        Input can be split on multiple lines as long as the line starts with
        a space.
        """
        proto = LDIFDriver()
        for line in (
            "dn: cn=foo,dc=ex",
            " ample,dc=com",
            "objectClass: a",
            "ob",
            " jectClass: b",
            "",
            ):
            proto.lineReceived(line.encode('ascii'))

        self.failUnlessEqual(len(proto.listOfCompleted), 1)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'a', b'b'])
        self.failUnlessEqual(proto.listOfCompleted, [])

    def testCaseInsensitiveDN(self):
        """
        DN is case insensitive.
        """
        proto = LDIFDriver()
        proto.dataReceived(
b"""version: 1
dN: cn=foo, dc=example, dc=com
cn: foo

DN: cn=bar, dc=example, dc=com
cn: bar

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 2)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'foo'])

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=bar,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'bar'])

        self.failUnlessEqual(proto.listOfCompleted, [])


    def testCaseInsensitiveAttributeTypes(self):
        """
        The attribute description (name/types) is case insensitive, while
        values are case sensitives.
        """
        proto = LDIFDriver()
        proto.dataReceived(b"""\
dn: cn=foo,dc=example,dc=com
objectClass: a
obJeCtClass: b
cn: foo
avalue: a
aValUe: B

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 1)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'a', b'b'])
        self.failUnlessEqual(o[b'CN'], [b'foo'])
        self.failUnlessEqual(o[b'aValue'], [b'a', b'B'])

        self.failUnlessEqual(proto.listOfCompleted, [])

    def testVersion1(self):
        proto = LDIFDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 1)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'a', b'b'])
        self.failUnlessEqual(o[b'aValue'], [b'a', b'b'])
        self.failUnlessEqual(o[b'bValue'], [b'c'])

    def testVersionInvalid(self):
        proto = LDIFDriver()
        self.assertRaises(ldifprotocol.LDIFVersionNotANumberError,
                          proto.dataReceived,
                          b"""\
version: junk
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

""")

    def testVersion2(self):
        proto = LDIFDriver()
        self.assertRaises(ldifprotocol.LDIFUnsupportedVersionError,
                          proto.dataReceived,
                          b"""\
version: 2
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

""")

    def testNoSpaces(self):
        proto = LDIFDriver()
        proto.dataReceived(b"""\
dn:cn=foo,dc=example,dc=com
objectClass:a
obJeCtClass:b
cn:foo
avalue:a
aValUe:b

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 1)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'objectClass'], [b'a', b'b'])
        self.failUnlessEqual(o[b'CN'], [b'foo'])
        self.failUnlessEqual(o[b'aValue'], [b'a', b'b'])

        self.failUnlessEqual(proto.listOfCompleted, [])

    def testTruncatedFailure(self):
        proto = LDIFDriver()
        proto.dataReceived(b"""\
version: 1
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c
""")

        self.failUnlessEqual(len(proto.listOfCompleted), 0)

        self.assertRaises(ldifprotocol.LDIFTruncatedError,
                          proto.connectionLost)

    def testComments(self):
        """
        Comments can be placed anywhere.
        """
        proto = LDIFDriver()
        proto.dataReceived(
b"""# One comment here.
version: 1
# After comment.
dn: cn=foo, dc=example, dc=com
# Another one comment here.
cn: foo

# More comments
dn: cn=bar, dc=example, dc=com
cn: bar

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 2)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'foo'])

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=bar,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'bar'])

        self.failUnlessEqual(proto.listOfCompleted, [])


    def testMoreEmptyLinesBetweenEntries(self):
        """
        It accept multiple lines between entries.
        """
        proto = LDIFDriver()
        proto.dataReceived(
b"""version: 1
dn: cn=foo, dc=example, dc=com
cn: foo



dn: cn=bar, dc=example, dc=com
cn: bar

""")

        self.failUnlessEqual(len(proto.listOfCompleted), 2)

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=foo,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'foo'])

        o = proto.listOfCompleted.pop(0)
        self.failUnlessEqual(o.dn.getText(), u'cn=bar,dc=example,dc=com')
        self.failUnlessEqual(o[b'CN'], [b'bar'])

        self.failUnlessEqual(proto.listOfCompleted, [])


    def testStartWithSpace(self):
        """
        It fails to parse if a line start with a space but is not a
        continuation of a previous line.
        """
        proto = LDIFDriver()
        with self.assertRaises(ldifprotocol.LDIFEntryStartsWithSpaceError):
            proto.dataReceived(
b"""version: 1
dn: cn=foo, dc=example, dc=com
cn: foo

 dn: cn=bar, dc=example, dc=com
cn: bar

""")


    def testEntryStartWithoutDN(self):
        """
        It fails to parse the entry does not start with DN.
        """
        proto = LDIFDriver()
        with self.assertRaises(ldifprotocol.LDIFEntryStartsWithNonDNError):
            proto.dataReceived(
b"""version: 1
cn: cn=foo, dc=example, dc=com
other: foo

""")


    def testAttributeValueFromURL(self):
        """
        Getting attribute values from URL is not supported.
        """
        proto = LDIFDriver()
        with self.assertRaises(NotImplementedError):
            proto.dataReceived(
b"""version: 1
dn: cn=foo, dc=example, dc=com
cn:< file:///path/to/data 

""")



class RFC2849_Examples(unittest.TestCase):
    examples = [
        ( b"""Example 1: An simple LDAP file with two entries""",
          b"""\
version: 1
dn: cn=Barbara Jensen, ou=Product Development, dc=airius, dc=com
objectclass: top
objectclass: person
objectclass: organizationalPerson
cn: Barbara Jensen
cn: Barbara J Jensen
cn: Babs Jensen
sn: Jensen
uid: bjensen
telephonenumber: +1 408 555 1212
description: A big sailing fan.

dn: cn=Bjorn Jensen, ou=Accounting, dc=airius, dc=com
objectclass: top
objectclass: person
objectclass: organizationalPerson
cn: Bjorn Jensen
sn: Jensen
telephonenumber: +1 408 555 1212

""",
          [ ( b'cn=Barbara Jensen,ou=Product Development,dc=airius,dc=com',
              { b'objectClass': [b'top', b'person', b'organizationalPerson'],
                b'cn': [b'Barbara Jensen',
                       b'Barbara J Jensen',
                       b'Babs Jensen'],
                b'sn': [b'Jensen'],
                b'uid': [b'bjensen'],
                b'telephonenumber': [b'+1 408 555 1212'],
                b'description': [b'A big sailing fan.'],
                }),

            ( b'cn=Bjorn Jensen,ou=Accounting,dc=airius,dc=com',
              {  b'objectClass': [b'top', b'person', b'organizationalPerson'],
                 b'cn': [b'Bjorn Jensen'],
                 b'sn': [b'Jensen'],
                 b'telephonenumber': [b'+1 408 555 1212'],
                 }),
            ]),

        ( b"""Example 2: A file containing an entry with a folded attribute value""",
          b"""\
version: 1
dn:cn=Barbara Jensen, ou=Product Development, dc=airius, dc=com
objectclass:top
objectclass:person
objectclass:organizationalPerson
cn:Barbara Jensen
cn:Barbara J Jensen
cn:Babs Jensen
sn:Jensen
uid:bjensen
telephonenumber:+1 408 555 1212
description:Babs is a big sailing fan, and travels extensively in sea
 rch of perfect sailing conditions.
title:Product Manager, Rod and Reel Division

""",
          [ ( b'cn=Barbara Jensen, ou=Product Development, dc=airius, dc=com',
              { b'objectclass': [b'top', b'person', b'organizationalPerson'],
                b'cn': [b'Barbara Jensen', b'Barbara J Jensen', b'Babs Jensen'],
                b'sn': [b'Jensen'],
                b'uid': [b'bjensen'],
                b'telephonenumber': [b'+1 408 555 1212'],
                b'description': [b'Babs is a big sailing fan, and travels extensively in search of perfect sailing conditions.'],
                b'title': [b'Product Manager, Rod and Reel Division'],
                }),
            ]),

        ( b"""Example 3: A file containing a base-64-encoded value""",
          b"""\
version: 1
dn: cn=Gern Jensen, ou=Product Testing, dc=airius, dc=com
objectclass: top
objectclass: person
objectclass: organizationalPerson
cn: Gern Jensen
cn: Gern O Jensen
sn: Jensen
uid: gernj
telephonenumber: +1 408 555 1212
description:: V2hhdCBhIGNhcmVmdWwgcmVhZGVyIHlvdSBhcmUhICBUaGlzIHZhbHVlIGlzIGJhc2UtNjQtZW5jb2RlZCBiZWNhdXNlIGl0IGhhcyBhIGNvbnRyb2wgY2hhcmFjdGVyIGluIGl0IChhIENSKS4NICBCeSB0aGUgd2F5LCB5b3Ugc2hvdWxkIHJlYWxseSBnZXQgb3V0IG1vcmUu

""",
          [ ( b'cn=Gern Jensen, ou=Product Testing, dc=airius, dc=com',
              { b'objectclass': [b'top', b'person', b'organizationalPerson'],
                b'cn': [b'Gern Jensen', b'Gern O Jensen'],
                b'sn': [b'Jensen'],
                b'uid': [b'gernj'],
                b'telephonenumber': [b'+1 408 555 1212'],
                b'description': [b'What a careful reader you are!  This value is base-64-encoded because it has a control character in it (a CR).\r  By the way, you should really get out more.'],
                }),
            ]),

        ]

    def testExamples(self):
        for name, data, expected in self.examples:
            proto = LDIFDriver()
            proto.dataReceived(data)

            self.failUnlessEqual(len(proto.listOfCompleted), len(expected))

            for dn, attr in expected:
                o = proto.listOfCompleted.pop(0)
                self.failUnlessEqual(o.dn, distinguishedname.DistinguishedName(dn))

                got = set([x.lower() for x in o.keys()])
                want = set([x.lower() for x in attr.keys()])
                self.failUnlessEqual(got, want)

                for k, v in attr.items():
                    self.failUnlessEqual(o[k], v)

            self.failUnlessEqual(proto.listOfCompleted, [])

"""
TODO more tests from RFC2849:

Example 4: A file containing an entries with UTF-8-encoded attribute
values, including language tags.  Comments indicate the contents
of UTF-8-encoded attributes and distinguished names.

version: 1
dn:: b3U95Za25qWt6YOoLG89QWlyaXVz
# dn:: ou=<JapaneseOU>,o=Airius
objectclass: top
objectclass: organizationalUnit
ou:: 5Za25qWt6YOo
# ou:: <JapaneseOU>
ou;lang-ja:: 5Za25qWt6YOo
# ou;lang-ja:: <JapaneseOU>
ou;lang-ja;phonetic:: 44GI44GE44GO44KH44GG44G2

# ou;lang-ja:: <JapaneseOU_in_phonetic_representation>
ou;lang-en: Sales
description: Japanese office

dn:: dWlkPXJvZ2FzYXdhcmEsb3U95Za25qWt6YOoLG89QWlyaXVz
# dn:: uid=<uid>,ou=<JapaneseOU>,o=Airius
userpassword: {SHA}O3HSv1MusyL4kTjP+HKI5uxuNoM=
objectclass: top
objectclass: person
objectclass: organizationalPerson
objectclass: inetOrgPerson
uid: rogasawara
mail: rogasawara@airius.co.jp
givenname;lang-ja:: 44Ot44OJ44OL44O8
# givenname;lang-ja:: <JapaneseGivenname>
sn;lang-ja:: 5bCP56yg5Y6f
# sn;lang-ja:: <JapaneseSn>
cn;lang-ja:: 5bCP56yg5Y6fIOODreODieODi+ODvA==
# cn;lang-ja:: <JapaneseCn>
title;lang-ja:: 5Za25qWt6YOoIOmDqOmVtw==
# title;lang-ja:: <JapaneseTitle>
preferredlanguage: ja
givenname:: 44Ot44OJ44OL44O8
# givenname:: <JapaneseGivenname>
sn:: 5bCP56yg5Y6f
# sn:: <JapaneseSn>
cn:: 5bCP56yg5Y6fIOODreODieODi+ODvA==
# cn:: <JapaneseCn>
title:: 5Za25qWt6YOoIOmDqOmVtw==
# title:: <JapaneseTitle>
givenname;lang-ja;phonetic:: 44KN44Gp44Gr44O8
# givenname;lang-ja;phonetic::
<JapaneseGivenname_in_phonetic_representation_kana>
sn;lang-ja;phonetic:: 44GK44GM44GV44KP44KJ
# sn;lang-ja;phonetic:: <JapaneseSn_in_phonetic_representation_kana>
cn;lang-ja;phonetic:: 44GK44GM44GV44KP44KJIOOCjeOBqeOBq+ODvA==
# cn;lang-ja;phonetic:: <JapaneseCn_in_phonetic_representation_kana>
title;lang-ja;phonetic:: 44GI44GE44GO44KH44GG44G2IOOBtuOBoeOCh+OBhg==
# title;lang-ja;phonetic::
# <JapaneseTitle_in_phonetic_representation_kana>
givenname;lang-en: Rodney
sn;lang-en: Ogasawara
cn;lang-en: Rodney Ogasawara
title;lang-en: Sales, Director
"""

"""
Example 5: A file containing a reference to an external file

version: 1
dn: cn=Horatio Jensen, ou=Product Testing, dc=airius, dc=com
objectclass: top
objectclass: person
objectclass: organizationalPerson
cn: Horatio Jensen

cn: Horatio N Jensen
sn: Jensen
uid: hjensen
telephonenumber: +1 408 555 1212
jpegphoto:< file:///usr/local/directory/photos/hjensen.jpg
"""

"""
Example 6: A file containing a series of change records and comments

version: 1
# Add a new entry
dn: cn=Fiona Jensen, ou=Marketing, dc=airius, dc=com
changetype: add
objectclass: top
objectclass: person
objectclass: organizationalPerson
cn: Fiona Jensen
sn: Jensen
uid: fiona
telephonenumber: +1 408 555 1212
jpegphoto:< file:///usr/local/directory/photos/fiona.jpg

# Delete an existing entry
dn: cn=Robert Jensen, ou=Marketing, dc=airius, dc=com
changetype: delete

# Modify an entry's relative distinguished name
dn: cn=Paul Jensen, ou=Product Development, dc=airius, dc=com
changetype: modrdn
newrdn: cn=Paula Jensen
deleteoldrdn: 1

# Rename an entry and move all of its children to a new location in
# the directory tree (only implemented by LDAPv3 servers).
dn: ou=PD Accountants, ou=Product Development, dc=airius, dc=com
changetype: modrdn
newrdn: ou=Product Development Accountants
deleteoldrdn: 0
newsuperior: ou=Accounting, dc=airius, dc=com

# Modify an entry: add an additional value to the postaladdress
# attribute, completely delete the description attribute, replace
# the telephonenumber attribute with two values, and delete a specific
# value from the facsimiletelephonenumber attribute
dn: cn=Paula Jensen, ou=Product Development, dc=airius, dc=com
changetype: modify
add: postaladdress
postaladdress: 123 Anystreet $ Sunnyvale, CA $ 94086
-
delete: description
-
replace: telephonenumber
telephonenumber: +1 408 555 1234
telephonenumber: +1 408 555 5678
-
delete: facsimiletelephonenumber
facsimiletelephonenumber: +1 408 555 9876
-

# Modify an entry: replace the postaladdress attribute with an empty
# set of values (which will cause the attribute to be removed), and
# delete the entire description attribute. Note that the first will
# always succeed, while the second will only succeed if at least
# one value for the description attribute is present.
dn: cn=Ingrid Jensen, ou=Product Support, dc=airius, dc=com
changetype: modify
replace: postaladdress
-
delete: description
-
"""

"""
Example 7: An LDIF file containing a change record with a control
version: 1
# Delete an entry. The operation will attach the LDAPv3
# Tree Delete Control defined in [9]. The criticality
# field is "true" and the controlValue field is
# absent, as required by [9].
dn: ou=Product Development, dc=airius, dc=com
control: 1.2.840.113556.1.4.805 true
changetype: delete

"""



# Copyright (C) 2001 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""LDAP protocol client"""

from ldaptor.protocols import pureldap, pureber
from ldaptor.protocols.ldap import ldaperrors

from twisted.python import log
from twisted.internet import protocol, defer, ssl, reactor


class LDAPClientConnectionLostException(ldaperrors.LDAPException):
    def toWire(self):
        return b'Connection lost'


class LDAPStartTLSBusyError(ldaperrors.LDAPOperationsError):
    def __init__(self, onwire, message=None):
        self.onwire = onwire
        ldaperrors.LDAPOperationsError.__init__(self, message=message)

    def toWire(self):
        return b'Cannot STARTTLS while operations on wire: %r' % self.onwire


class LDAPStartTLSInvalidResponseName(ldaperrors.LDAPException):
    def __init__(self, responseName):
        self.responseName = responseName
        ldaperrors.LDAPException.__init__(self)

    def toWire(self):
        return b'Invalid responseName in STARTTLS response: %r' % (self.responseName, )


class LDAPClient(protocol.Protocol):
    """An LDAP client"""
    debug = False

    def __init__(self):
        self.onwire = {}
        self.buffer = b''
        self.connected = None

    berdecoder = pureldap.LDAPBERDecoderContext_TopLevel(
        inherit=pureldap.LDAPBERDecoderContext_LDAPMessage(
            fallback=pureldap.LDAPBERDecoderContext(fallback=pureber.BERDecoderContext()),
            inherit=pureldap.LDAPBERDecoderContext(fallback=pureber.BERDecoderContext())))

    def dataReceived(self, recd):
        self.buffer += recd
        while 1:
            try:
                o, bytes = pureber.berDecodeObject(self.berdecoder, self.buffer)
            except pureber.BERExceptionInsufficientData:
                o, bytes = None, 0
            self.buffer = self.buffer[bytes:]
            if not o:
                break
            self.handle(o)

    def connectionMade(self):
        """TCP connection has opened"""
        self.connected = 1

    def connectionLost(self, reason=protocol.connectionDone):
        """Called when TCP connection has been lost"""
        self.connected = 0
        # notify handlers of operations in flight
        while self.onwire:
            k, v = self.onwire.popitem()
            d, _, _, _, _ = v
            d.errback(reason)

    def _send(self, op, controls=None):
        if not self.connected:
            raise LDAPClientConnectionLostException()
        msg = pureldap.LDAPMessage(op, controls=controls)
        if self.debug:
            log.msg('C->S %s' % repr(msg))
        assert msg.id not in self.onwire
        return msg

    def send(self, op, controls=None):
        """
        Send an LDAP operation to the server.
        @param op: the operation to send
        @type op: LDAPProtocolRequest
        @param controls: Any controls to be included in the request.
        @type controls: LDAPControls
        @return: the response from server
        @rtype: Deferred LDAPProtocolResponse
        """
        msg = self._send(op, controls=controls)
        assert op.needs_answer
        d = defer.Deferred()
        self.onwire[msg.id] = (d, False, None, None, None)
        self.transport.write(msg.toWire())
        return d

    def send_multiResponse(self, op, handler, *args, **kwargs):
        """
        Send an LDAP operation to the server, expecting one or more
        responses.

        If `handler` is provided, it will receive a LDAP response as
        its first argument. The Deferred returned by this function will
        never fire.

        If `handler` is not provided, the Deferred returned by this
        function will fire with the final LDAP response.

        @param op: the operation to send
        @type op: LDAPProtocolRequest
        @param handler: a callable that will be called for each
        response. It should return a boolean, whether this was the
        final response.
        @param args: positional arguments to pass to handler
        @param kwargs: keyword arguments to pass to handler
        @return: the result from the first handler as a deferred that
        completes when the first response has been received
        @rtype: Deferred LDAPProtocolResponse
        """
        msg = self._send(op)
        assert op.needs_answer
        d = defer.Deferred()
        self.onwire[msg.id] = (d, False, handler, args, kwargs)
        self.transport.write(msg.toWire())
        return d

    def send_multiResponse_ex(self, op, controls=None, handler=None, *args, **kwargs):
        """
        Send an LDAP operation to the server, expecting one or more
        responses.

        If `handler` is provided, it will receive a LDAP response *and*
        response controls as its first 2 arguments. The Deferred returned
        by this function will never fire.

        If `handler` is not provided, the Deferred returned by this
        function will fire with a tuple of the first LDAP response
        and any associated response controls.

        @param op: the operation to send
        @type op: LDAPProtocolRequest
        @param controls: LDAP controls to send with the message.
        @type controls: LDAPControls
        @param handler: a callable that will be called for each
        response. It should return a boolean, whether this was the
        final response.
        @param args: positional arguments to pass to handler
        @param kwargs: keyword arguments to pass to handler
        @return: the result from the last handler as a deferred that
        completes when the last response has been received
        @rtype: Deferred LDAPProtocolResponse
        """
        msg = self._send(op, controls=controls)
        assert op.needs_answer
        d = defer.Deferred()
        self.onwire[msg.id] = (d, True, handler, args, kwargs)
        self.transport.write(msg.toWire())
        return d

    def send_noResponse(self, op, controls=None):
        """
        Send an LDAP operation to the server, with no response
        expected.

        @param op: the operation to send
        @type op: LDAPProtocolRequest
        """
        msg = self._send(op, controls=controls)
        assert not op.needs_answer
        self.transport.write(msg.toWire())

    def unsolicitedNotification(self, msg):
        log.msg("Got unsolicited notification: %s" % repr(msg))

    def handle(self, msg):
        assert isinstance(msg.value, pureldap.LDAPProtocolResponse)
        if self.debug:
            log.msg('C<-S %s' % repr(msg))

        if msg.id == 0:
            self.unsolicitedNotification(msg.value)
        else:
            d, return_controls, handler, args, kwargs = self.onwire[msg.id]

            if handler is None:
                assert (args is None) or (args == ())
                assert (kwargs is None) or (kwargs == {})
                if return_controls:
                    d.callback((msg.value, msg.controls))
                else:
                    d.callback(msg.value)
                del self.onwire[msg.id]
            else:
                assert args is not None
                assert kwargs is not None
                # Return true to mark request as fully handled
                if return_controls:
                    if handler(msg.value, msg.controls, *args, **kwargs):
                        del self.onwire[msg.id]
                else:
                    if handler(msg.value, *args, **kwargs):
                        del self.onwire[msg.id]

    def bind(self, dn='', auth=''):
        """
        @depreciated: Use e.bind(auth).

        @todo: Remove this method when there are no callers.
        """
        if not self.connected:
            raise LDAPClientConnectionLostException()
        else:
            r = pureldap.LDAPBindRequest(dn=dn, auth=auth)
            d = self.send(r)
            d.addCallback(self._handle_bind_msg)
        return d

    def _handle_bind_msg(self, msg):
        assert isinstance(msg, pureldap.LDAPBindResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)
        return (msg.matchedDN, msg.serverSaslCreds)

    def unbind(self):
        if not self.connected:
            raise Exception("Not connected (TODO)")  # TODO make this a real object
        r = pureldap.LDAPUnbindRequest()
        self.send_noResponse(r)
        self.transport.loseConnection()

    def _cbStartTLS(self, msg, ctx):
        assert isinstance(msg, pureldap.LDAPExtendedResponse)
        assert msg.referral is None  # TODO
        if msg.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(msg.resultCode, msg.errorMessage)

        if (msg.responseName is not None) and \
                (msg.responseName != pureldap.LDAPStartTLSResponse.oid):
            raise LDAPStartTLSInvalidResponseName(msg.responseName)

        self.transport.startTLS(ctx)
        return self

    def startTLS(self, ctx=None):
        """
        Start Transport Layer Security.

        It is the callers responsibility to make sure other things
        are not happening at the same time.

        @todo: server hostname check, see rfc2830 section 3.6.
        @return: a deferred that will complete when the TLS handshake is
        complete.
        """
        if ctx is None:
            ctx = ssl.ClientContextFactory()
        # we always delay by one event loop iteration to make
        # sure the previous handler has exited and self.onwire
        # has been cleaned up
        d = defer.Deferred()
        d.addCallback(self._startTLS)
        reactor.callLater(0, d.callback, ctx)
        return d

    def _startTLS(self, ctx):
        if not self.connected:
            raise LDAPClientConnectionLostException()
        elif self.onwire:
            raise LDAPStartTLSBusyError(self.onwire)
        else:
            op = pureldap.LDAPStartTLSRequest()
            d = self.send(op)
            d.addCallback(self._cbStartTLS, ctx)
            return d



# Copyright (C) 2001 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Test cases for ldaptor.protocols.pureber module.
"""
import six
from twisted.trial import unittest

from ldaptor.protocols import pureber


def s(*l):
    """Join all members of list to a byte string. Integer members are converted to bytes"""
    r = b''
    for e in l:
        e = six.int2byte(e)
        r = r + e
    return r


def l(s):
    """Split a byte string to ord's of chars."""
    return [six.byte2int([x]) for x in s]


class BerLengths(unittest.TestCase):
    knownValues=(
        (0, [0]),
        (1, [1]),
        (100, [100]),
        (126, [126]),
        (127, [127]),
        (128, [0x80|1, 128]),
        (129, [0x80|1, 129]),
        (255, [0x80|1, 255]),
        (256, [0x80|2, 1, 0]),
        (257, [0x80|2, 1, 1]),
        (65535, [0x80|2, 0xFF, 0xFF]),
        (65536, [0x80|3, 0x01, 0x00, 0x00]),
        (256**127-1, [0x80|127]+127*[0xFF]),
        )

    def testToBER(self):
        for integer, encoded in self.knownValues:
            got = pureber.int2berlen(integer)
            got = bytes(got)
            got = l(got)
            self.assertEqual(got, encoded)

    def testFromBER(self):
        for integer, encoded in self.knownValues:
            m=s(*encoded)
            got, bytes = pureber.berDecodeLength(m)
            self.assertEqual(bytes, len(m))
            self.assertEqual(got, integer)

    def testPartialBER(self):
        m = bytes(pureber.int2berlen(3*256))
        self.assertEqual(3, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:1])

        m = bytes(pureber.int2berlen(256**100-1))
        self.assertEqual(101, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:100])


class BERBaseTests(unittest.TestCase):
    """
    Unit tests for generic BERBase.
    """
    valuesToTest=(
        (pureber.BERBase, []),
        (pureber.BERInteger, [0]),
        (pureber.BERInteger, [1]),
        (pureber.BERInteger, [4000]),
        (pureber.BERSequence, [[pureber.BERInteger(1000), pureber.BERInteger(2000)]]),
        (pureber.BERSequence, [[pureber.BERInteger(2000), pureber.BERInteger(1000)]]),
        (pureber.BEROctetString, ["foo"]),
        (pureber.BEROctetString, ["b"+chr(0xe4)+chr(0xe4)]),
        )

    def testEquality(self):
        """
        BER objects equal BER objects with same type and content
        """
        for class_, args in self.valuesToTest:
            x=class_(*args)
            y=class_(*args)
            self.assertEqual(x, x)
            self.assertEqual(x, y)

    def testInequalityWithBER(self):
        """
        BER objects do not equal BER objects with different type or content
        """
        for i in six.moves.range(len(self.valuesToTest)):
            for j in six.moves.range(len(self.valuesToTest)):
                if i!=j:
                    i_class, i_args = self.valuesToTest[i]
                    j_class, j_args = self.valuesToTest[j]
                    x=i_class(*i_args)
                    y=j_class(*j_args)
                    self.assertNotEqual(x, y)

    def testInequalityWithNonBER(self):
        """
        BER objects are not equal with non-BER objects.
        """
        sut = pureber.BERInteger([0])

        self.assertFalse(0 == sut)
        self.assertNotEqual(0, sut)

    def testHashEquality(self):
        """
        Objects which are equal have the same hash.
        """
        for klass, arguments in self.valuesToTest:
            first = klass(*arguments)
            second = klass(*arguments)
            self.assertEqual(hash(first), hash(second))


class BERIntegerKnownValues(unittest.TestCase):
    knownValues=(
        (0, [0x02, 0x01, 0]),
        (1, [0x02, 0x01, 1]),
        (2, [0x02, 0x01, 2]),
        (125, [0x02, 0x01, 125]),
        (126, [0x02, 0x01, 126]),
        (127, [0x02, 0x01, 127]),
        (-1, [0x02, 0x01, 256-1]),
        (-2, [0x02, 0x01, 256-2]),
        (-3, [0x02, 0x01, 256-3]),
        (-126, [0x02, 0x01, 256-126]),
        (-127, [0x02, 0x01, 256-127]),
        (-128, [0x02, 0x01, 256-128]),
        (-129, [0x02, 0x02, 256-1, 256-129]),
        (128, [0x02, 0x02, 0, 128]),
        (256, [0x02, 0x02, 1, 0]),
        )

    def testToBERIntegerKnownValues(self):
        """BERInteger(n).toWire() should give known result with known input"""
        for integer, encoded in self.knownValues:
            result = pureber.BERInteger(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBERIntegerKnownValues(self):
        """BERInteger(encoded="...") should give known result with known input"""
        for integer, encoded in self.knownValues:
            m=s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERInteger)
            result = result.value
            self.assertEqual(integer, result)

    def testPartialBERIntegerEncodings(self):
        """BERInteger(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERInteger(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))


class BERIntegerSanityCheck(unittest.TestCase):
    def testSanity(self):
        """BERInteger(encoded=BERInteger(n)).value==n for -1000..1000"""
        for n in range(-1000, 1001, 10):
            encoded = pureber.BERInteger(n).toWire()
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), encoded)
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BERInteger)
            result = result.value
            self.assertEqual(n, result)


class ObjectWithToWireMethod(object):
    def toWire(self):
        return b"bar"


class TestBEROctetString(unittest.TestCase):
    """
    Unit tests for BEROctetString.
    """
    knownValues=(
        ("", [0x04, 0]),
        ("foo", [0x04, 3]+l(b"foo")),
        (100 * "x", [0x04, 100]+l(100 * b"x")),
        (ObjectWithToWireMethod(), [0x04, 3]+l(b"bar")),
        )

    def testToBEROctetStringKnownValues(self):
        """BEROctetString(n).toWire() should give known result with known input"""
        for st, encoded in self.knownValues:
            result = pureber.BEROctetString(st)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBEROctetStringKnownValues(self):
        """BEROctetString(encoded="...") should give known result with known input"""
        for st, encoded in self.knownValues:
            m=s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BEROctetString)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testPartialBEROctetStringEncodings(self):
        """BEROctetString(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BEROctetString("x").toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))

    def testSanity(self):
        """BEROctetString(encoded=BEROctetString(n*'x')).value==n*'x' for some values of n"""
        for n in 0,1,2,3,4,5,6,100,126,127,128,129,1000,2000:
            encoded = pureber.BEROctetString(n * b'x').toWire()
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), encoded)
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BEROctetString)
            result = result.value
            self.assertEqual(n * b'x', result)


class BERNullKnownValues(unittest.TestCase):
    def testToBERNullKnownValues(self):
        """BERNull().toWire() should give known result"""
        result = pureber.BERNull()
        result = result.toWire()
        result = l(result)
        self.assertEqual([0x05, 0x00], result)

    def testFromBERNullKnownValues(self):
        """BERNull(encoded="...") should give known result with known input"""
        encoded=[0x05, 0x00]
        m=s(*encoded)
        result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
        self.assertEqual(bytes, len(m))
        self.assertIsInstance(result, pureber.BERNull)
        self.assertEqual(0x05, result.tag)

    def testPartialBERNullEncodings(self):
        """BERNull(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERNull().toWire()
        self.assertEqual(2, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))


class BERBooleanKnownValues(unittest.TestCase):
    knownValues=(
        (0, [0x01, 0x01, 0], 0),
        (1, [0x01, 0x01, 0xFF], 0xFF),
        (2, [0x01, 0x01, 0xFF], 0xFF),
        (125, [0x01, 0x01, 0xFF], 0xFF),
        (126, [0x01, 0x01, 0xFF], 0xFF),
        (127, [0x01, 0x01, 0xFF], 0xFF),
        (-1, [0x01, 0x01, 0xFF], 0xFF),
        (-2, [0x01, 0x01, 0xFF], 0xFF),
        (-3, [0x01, 0x01, 0xFF], 0xFF),
        (-126, [0x01, 0x01, 0xFF], 0xFF),
        (-127, [0x01, 0x01, 0xFF], 0xFF),
        (-128, [0x01, 0x01, 0xFF], 0xFF),
        (-129, [0x01, 0x01, 0xFF], 0xFF),
        (-9999, [0x01, 0x01, 0xFF], 0xFF),
        (128, [0x01, 0x01, 0xFF], 0xFF),
        (255, [0x01, 0x01, 0xFF], 0xFF),
        (256, [0x01, 0x01, 0xFF], 0xFF),
        (9999, [0x01, 0x01, 0xFF], 0xFF),
        )

    def testToBERBooleanKnownValues(self):
        """BERBoolean(n).toWire() should give known result with known input"""
        for integer, encoded, dummy in self.knownValues:
            result = pureber.BERBoolean(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBERBooleanKnownValues(self):
        """BERBoolean(encoded="...") should give known result with known input"""
        for integer, encoded, canon in self.knownValues:
            m=s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERBoolean)
            result = result.value
            self.assertEqual(canon, result)

    def testPartialBERBooleanEncodings(self):
        """BERBoolean(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERBoolean(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))


class BEREnumeratedKnownValues(unittest.TestCase):
    knownValues=(
        (0, [0x0a, 0x01, 0]),
        (1, [0x0a, 0x01, 1]),
        (2, [0x0a, 0x01, 2]),
        (125, [0x0a, 0x01, 125]),
        (126, [0x0a, 0x01, 126]),
        (127, [0x0a, 0x01, 127]),
        (-1, [0x0a, 0x01, 256-1]),
        (-2, [0x0a, 0x01, 256-2]),
        (-3, [0x0a, 0x01, 256-3]),
        (-126, [0x0a, 0x01, 256-126]),
        (-127, [0x0a, 0x01, 256-127]),
        (-128, [0x0a, 0x01, 256-128]),
        (-129, [0x0a, 0x02, 256-1, 256-129]),
        (128, [0x0a, 0x02, 0, 128]),
        (256, [0x0a, 0x02, 1, 0]),
        )

    def testToBEREnumeratedKnownValues(self):
        """BEREnumerated(n).toWire() should give known result with known input"""
        for integer, encoded in self.knownValues:
            result = pureber.BEREnumerated(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBEREnumeratedKnownValues(self):
        """BEREnumerated(encoded="...") should give known result with known input"""
        for integer, encoded in self.knownValues:
            m=s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BEREnumerated)
            result = result.value
            self.assertEqual(integer, result)

    def testPartialBEREnumeratedEncodings(self):
        """BEREnumerated(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BEREnumerated(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))


class BEREnumeratedSanityCheck(unittest.TestCase):
    def testSanity(self):
        """BEREnumerated(encoded=BEREnumerated(n)).value==n for -1000..1000"""
        for n in range(-1000, 1001, 10):
            encoded = pureber.BEREnumerated(n).toWire()
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), encoded)
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BEREnumerated)
            result = result.value
            self.assertEqual(n, result)


class TestBERSequence(unittest.TestCase):
    """
    Unit test for BERSequence.
    """

    def testStringRepresentationEmpty(self):
        """
        It can return the byte string representation for empty sequence which
        is just the zero/null byte.
        """
        sut = pureber.BERSequence([])

        result = sut.toWire()

        self.assertEqual(b'0\x00', result)

    def testStringRepresentatinSmallInteger(self):
        """
        It can represent a sequence of a single integer which has a
        single byte value.
        """
        sut = pureber.BERSequence([pureber.BERInteger(2)])

        result = sut.toWire()

        self.assertEqual(b'0\x03\x02\x01\x02', result)

    def testStringRepresentatinLargerInteger(self):
        """
        It can represent a sequence of a single integer which has a
        multi bites value.
        """
        sut = pureber.BERSequence([pureber.BERInteger(128)])

        result = sut.toWire()

        self.assertEqual(b'0\x04\x02\x02\x00\x80', result)

    def testStringRepresentatinMultipleIntegers(self):
        """
        It can represent a sequence of multiple integer.
        """
        sut = pureber.BERSequence([
            pureber.BERInteger(3), pureber.BERInteger(128)])

        result = sut.toWire()

        self.assertEqual(b'0\x07\x02\x01\x03\x02\x02\x00\x80', result)

    def testDecodeValidInput(self):
        """
        It can be decoded from its bytes serialization.
        """
        knownValues=(
            ([], [0x30, 0x00]),
            ([pureber.BERInteger(2)], [0x30, 0x03, 0x02, 0x01, 2]),
            ([pureber.BERInteger(3)], [0x30, 0x03, 0x02, 0x01, 3]),
            ([pureber.BERInteger(128)], [0x30, 0x04, 0x02, 0x02, 0, 128]),
            ([
                pureber.BERInteger(2),
                pureber.BERInteger(3),
                pureber.BERInteger(128),
                ],
             [0x30, 0x0a] + [0x02, 0x01, 2] + [0x02, 0x01, 3] + [0x02, 0x02, 0, 128]),
            )

        for content, encoded in knownValues:
            m=s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERSequence)
            result = result.data
            self.assertEqual(len(content), len(result))
            for i in six.moves.range(len(content)):
                self.assertEqual(content[i], result[i])
            self.assertEqual(content, result)

    def testDecdeInvalidInput(self):
        """
        It raises BERExceptionInsufficientData when trying to decode from
        data which is not valid.
        """
        m = pureber.BERSequence([pureber.BERInteger(2)]).toWire()
        self.assertEqual(5, len(m))

        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:4])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:3])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:2])
        self.assertRaises(pureber.BERExceptionInsufficientData, pureber.berDecodeObject, pureber.BERDecoderContext(), m[:1])
        self.assertEqual((None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), ''))



#!/usr/bin/python

import six

from ldaptor.protocols import pureldap
from ldaptor._encoder import to_unicode


"""

RFC2254:

        filter     = "(" filtercomp ")"
        filtercomp = and / or / not / item
        and        = "&" filterlist
        or         = "|" filterlist
        not        = "!" filter
        filterlist = 1*filter
        item       = simple / present / substring / extensible
        simple     = attr filtertype value
        filtertype = equal / approx / greater / less
        equal      = "="
        approx     = "~="
        greater    = ">="
        less       = "<="
        extensible = attr [":dn"] [":" matchingrule] ":=" value
                     / [":dn"] ":" matchingrule ":=" value
        present    = attr "=*"
        substring  = attr "=" [initial] any [final]
        initial    = value
        any        = "*" *(value "*")
        final      = value
        attr       = AttributeDescription from Section 4.1.5 of [1]
        matchingrule = MatchingRuleId from Section 4.1.9 of [1]
        value      = AttributeValue from Section 4.1.6 of [1]
"""


class InvalidLDAPFilter(Exception):
    def __init__(self, msg, loc, text):
        Exception.__init__(self)
        self.msg = msg
        self.loc = loc
        self.text = text

    def __str__(self):
        return "Invalid LDAP filter: %s at point %d in %r" \
               % (self.msg, self.loc, self.text)


def parseExtensible(attr, s):
    raise NotImplementedError()



from pyparsing import Word, Literal, Optional, ZeroOrMore, Suppress, \
                       Group, Forward, OneOrMore, ParseException, \
                       CharsNotIn, Combine, StringStart, \
                       StringEnd, delimitedList

import string

filter_ = Forward()
attr = Word(string.ascii_letters,
            string.ascii_letters + string.digits + ';-',)
attr.leaveWhitespace()
attr.setName('attr')
hexdigits = Word(string.hexdigits, exact=2)
hexdigits.setName('hexdigits')
escaped = Suppress(Literal('\\')) + hexdigits
escaped.setName('escaped')


def _p_escaped(s, l, t):
    text = t[0]
    return chr(int(text, 16))


escaped.setParseAction(_p_escaped)
value = Combine(OneOrMore(CharsNotIn('*()\\\0') | escaped))
value.setName('value')
equal = Literal("=")
equal.setParseAction(lambda s, l, t: pureldap.LDAPFilter_equalityMatch)
approx = Literal("~=")
approx.setParseAction(lambda s, l, t: pureldap.LDAPFilter_approxMatch)
greater = Literal(">=")
greater.setParseAction(lambda s, l, t: pureldap.LDAPFilter_greaterOrEqual)
less = Literal("<=")
less.setParseAction(lambda s, l, t: pureldap.LDAPFilter_lessOrEqual)
filtertype = equal | approx | greater | less
filtertype.setName('filtertype')
simple = attr + filtertype + value
simple.leaveWhitespace()
simple.setName('simple')


def _p_simple(s, l, t):
    attr, filtertype, value = t
    return filtertype(attributeDesc=pureldap.LDAPAttributeDescription(attr),
                      assertionValue=pureldap.LDAPAssertionValue(value))


simple.setParseAction(_p_simple)
present = attr + "=*"
present.setParseAction(lambda s, l, t: pureldap.LDAPFilter_present(t[0]))
initial = value.copy()
initial.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_initial(t[0]))
initial.setName('initial')
any_value = value + Suppress(Literal("*"))
any_value.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_any(t[0]))
any = Suppress(Literal("*")) + ZeroOrMore(any_value)
any.setName('any')
final = value.copy()
final.setName('final')
final.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_final(t[0]))
substring = attr + Suppress(Literal("=")) + Group(Optional(initial) + any + Optional(final))
substring.setName('substring')


def _p_substring(s, l, t):
    attrtype, substrings = t
    return pureldap.LDAPFilter_substrings(
        type=attrtype,
        substrings=substrings)


substring.setParseAction(_p_substring)

keystring = Word(string.ascii_letters,
                 string.ascii_letters + string.digits + ';-')
keystring.setName('keystring')
numericoid = delimitedList(Word(string.digits), delim='.', combine=True)
numericoid.setName('numericoid')
oid = numericoid | keystring
oid.setName('oid')
matchingrule = oid.copy()
matchingrule.setName('matchingrule')

extensible_dn = Optional(":dn")


def _p_extensible_dn(s, l, t):
    return bool(t)


extensible_dn.setParseAction(_p_extensible_dn)

matchingrule_or_none = Optional(Suppress(":") + matchingrule)


def _p_matchingrule_or_none(s, l, t):
    if not t:
        return [None]
    else:
        return t[0]


matchingrule_or_none.setParseAction(_p_matchingrule_or_none)

extensible_attr = attr + extensible_dn + matchingrule_or_none + Suppress(":=") + value
extensible_attr.setName('extensible_attr')


def _p_extensible_attr(s, l, t):
    return list(t)


extensible_attr.setParseAction(_p_extensible_attr)

extensible_noattr = extensible_dn + Suppress(":") + matchingrule + Suppress(":=") + value
extensible_noattr.setName('extensible_noattr')


def _p_extensible_noattr(s, l, t):
    return [None] + list(t)


extensible_noattr.setParseAction(_p_extensible_noattr)

extensible = extensible_attr | extensible_noattr
extensible.setName('extensible')


def _p_extensible(s, l, t):
    attr, dn, matchingRule, value = t
    return pureldap.LDAPFilter_extensibleMatch(
        matchingRule=matchingRule,
        type=attr,
        matchValue=value,
        dnAttributes=dn)


extensible.setParseAction(_p_extensible)
item = simple ^ present ^ substring ^ extensible
item.setName('item')
item.leaveWhitespace()
not_ = Suppress(Literal('!')) + filter_
not_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_not(t[0]))
not_.setName('not')
filterlist = OneOrMore(filter_)
or_ = Suppress(Literal('|')) + filterlist
or_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_or(t))
or_.setName('or')
and_ = Suppress(Literal('&')) + filterlist
and_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_and(t))
and_.setName('and')
filtercomp = and_ | or_ | not_ | item
filtercomp.setName('filtercomp')
filter_ << (Suppress(Literal('(').leaveWhitespace())
            + filtercomp
            + Suppress(Literal(')').leaveWhitespace()))
filter_.setName('filter')
filtercomp.leaveWhitespace()
filter_.leaveWhitespace()

toplevel = (StringStart().leaveWhitespace()
            + filter_
            + StringEnd().leaveWhitespace())
toplevel.leaveWhitespace()
toplevel.setName('toplevel')


def parseFilter(s):
    """
    Converting source string to pureldap.LDAPFilter

    Source string is converted to unicode for Python 3
    as pyparsing cannot parse Python 3 byte strings with
    the rules declared in this module.
    """
    s = to_unicode(s) if not six.PY2 else s
    try:
        x = toplevel.parseString(s)
    except ParseException as e:
        raise InvalidLDAPFilter(e.msg,
                                  e.loc,
                                  e.line)
    assert len(x) == 1
    return x[0]


maybeSubString_value = Combine(OneOrMore(CharsNotIn('*\\\0') | escaped))

maybeSubString_simple = maybeSubString_value.copy()


def _p_maybeSubString_simple(s, l, t):
    return (lambda attr:
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureldap.LDAPAttributeDescription(attr),
                assertionValue=pureldap.LDAPAssertionValue(t[0])))


maybeSubString_simple.setParseAction(_p_maybeSubString_simple)

maybeSubString_present = Literal("*")


def _p_maybeSubString_present(s, l, t):
    return (lambda attr:
            pureldap.LDAPFilter_present(attr))


maybeSubString_present.setParseAction(_p_maybeSubString_present)

maybeSubString_substring = Optional(initial) + any + Optional(final)


def _p_maybeSubString_substring(s, l, t):
    return (lambda attr:
            pureldap.LDAPFilter_substrings(
                type=attr,
                substrings=t))


maybeSubString_substring.setParseAction(_p_maybeSubString_substring)

maybeSubString = (maybeSubString_simple
                  ^ maybeSubString_present
                  ^ maybeSubString_substring
                  )


def parseMaybeSubstring(attrType, s):
    try:
        x = maybeSubString.parseString(s)
    except ParseException as e:
        raise InvalidLDAPFilter(e.msg,
                                  e.loc,
                                  e.line)
    assert len(x) == 1
    fn = x[0]
    return fn(attrType)


if __name__ == '__main__':
    import sys

    for filt in sys.argv[1:]:
        print(repr(parseFilter(filt)))
        print()



"""
Tests for the code from docs/source/example.
"""
from __future__ import absolute_import
import os
import sys

from twisted.test import proto_helpers
from twisted.trial import unittest

from ldaptor import inmemory
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols import pureldap

# We inject the examples so that we can import them.
# There is no cleanup, so this is leaving side effects.
sys.path.append(os.path.abspath("docs/source/examples"))
import ldaptor_with_upn_bind


class LDAPServerWithUPNBind(unittest.TestCase):
    """
    Tests for docs/source/examples/ldaptor_with_upn_bind.py
    """

    def setUp(self):
        self.root = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn='dc=example,dc=com',
            attributes={'dc': 'example'})
        self.user = self.root.addChild(
            rdn=b'cn=bob',
            attributes={
                'objectClass': ['a', 'b'],
                # Hash is for "secret".
                'userPassword': [b'{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8='],
                'userPrincipalName': ['bob@ad.example.com'],
            })

        server = ldaptor_with_upn_bind.LDAPServerWithUPNBind()
        server.factory = self.root
        server.transport = proto_helpers.StringTransport()
        server.connectionMade()
        self.server = server

    def checkSuccessfulBIND(self, bind_dn, password):
        """
        Do a BIND request and check that is succeeds.
        """
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn=bind_dn,
                    auth=password),
                id=4).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0,
                    matchedDN='cn=bob,dc=example,dc=com'),
                id=4).toWire()
        )

    def test_bindSuccessUPN(self):
        """
        It can authenticate based on the UPN.
        """
        self.checkSuccessfulBIND('bob@ad.example.com', b'secret')


    def test_bindSuccessDN(self):
        """
        It can still authenticate based on the normal DN.
        """
        self.checkSuccessfulBIND('cn=bob,dc=example,dc=com', b'secret')


    def test_bindBadPassword(self):
        """
        When password don't match the BIND fails.
        """
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn='bob@ad.example.com',
                    auth='invalid'),
                id=734).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode),
                id=734).toWire()
        )



"""Find an available uidNumber/gidNumber/other similar number."""
from __future__ import division

from ldaptor.protocols import pureldap

class freeNumberGuesser:
    def __init__(self, makeAGuess, min=None, max=None):
        self.makeAGuess=makeAGuess
        self.min=min
        if self.min is None:
            self.min=0
        self.max=max

    def startGuessing(self):
        d=self.makeAGuess(self.min)
        d.addCallback(self._nextGuess, self.min)
        return d

    def _nextGuess(self, found, lastGuess):
        if found:
            self.min=lastGuess
        else:
            self.max=lastGuess

        if self.max==self.min \
           or self.max==self.min+1:
            return self.max

        max=self.max
        if max is None:
            max=self.min+1000

        guess = (max + self.min) // 2
        d=self.makeAGuess(guess)
        d.addCallback(self._nextGuess, guess)
        return d

class ldapGuesser:
    def __init__(self, ldapObject, numberType):
        self.numberType=numberType
        self.ldapObject=ldapObject

    def guess(self, num):
        d=self.ldapObject.search(
            filterObject=pureldap.LDAPFilter_equalityMatch(
            attributeDesc=pureldap.LDAPAttributeDescription(value=self.numberType),
            assertionValue=pureldap.LDAPAssertionValue(value=str(num))),
            sizeLimit=1)
        d.addCallback(lambda results: len(results))
        return d

def getFreeNumber(ldapObject, numberType, min=None, max=None):
    g=freeNumberGuesser(ldapGuesser(ldapObject, numberType).guess,
                        min=min, max=max)
    return g.startGuessing()



import os
from zope.interface import Interface, implements
from twisted.internet import reactor, defer
from twisted.cred import portal, checkers
from nevow import rend, appserver, inevow, \
     stan, guard, loaders, flat
from formless import annotate, webform

from ldaptor.protocols.ldap import ldapclient, ldapsyntax, ldapconnector, \
     distinguishedname
from ldaptor import ldapfilter
from ldaptor.protocols import pureldap

class ILDAPConfig(Interface):
    """Addressbook configuration retrieval."""

    def getBaseDN(self):
        """Get the LDAP base DN, as a DistinguishedName."""

    def getServiceLocationOverrides(self):
        """
        Get the LDAP service location overrides, as a mapping of
        DistinguishedName to (host, port) tuples.
        """

class LDAPConfig(object):
    implements(ILDAPConfig)

    def __init__(self,
                 baseDN,
                 serviceLocationOverrides=None):
        self.baseDN = distinguishedname.DistinguishedName(baseDN)
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k,v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn]=v

    def getBaseDN(self):
        return self.baseDN

    def getServiceLocationOverrides(self):
        return self.serviceLocationOverrides

class LDAPSearchFilter(annotate.String):
    def coerce(self, *a, **kw):
        val = super(LDAPSearchFilter, self).coerce(*a, **kw)
        try:
            f = ldapfilter.parseFilter(val)
        except ldapfilter.InvalidLDAPFilter, e:
            raise annotate.InputError, \
                  "%r is not a valid LDAP search filter: %s" % (val, e)
        return f

class IAddressBookSearch(annotate.TypedInterface):
    search = LDAPSearchFilter(label="Search filter")

class CurrentSearch(object):
    implements(IAddressBookSearch, inevow.IContainer)
    search = None

    def child(self, context, name):
        if name == 'searchFilter':
            return self.search
        if name != 'results':
            return None
        config = context.locate(ILDAPConfig)

        c=ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d=c.connectAnonymously(config.getBaseDN(),
                               config.getServiceLocationOverrides())

        def _search(proto, base, searchFilter):
            baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=base)
            d=baseEntry.search(filterObject=searchFilter)
            return d

        d.addCallback(_search, config.getBaseDN(), self.search)
        return d

def LDAPFilterSerializer(original, context):
    return original.asText()

# TODO need to make this pretty some day.
for c in [
    pureldap.LDAPFilter_and,
    pureldap.LDAPFilter_or,
    pureldap.LDAPFilter_not,
    pureldap.LDAPFilter_substrings,
    pureldap.LDAPFilter_equalityMatch,
    pureldap.LDAPFilter_greaterOrEqual,
    pureldap.LDAPFilter_lessOrEqual,
    pureldap.LDAPFilter_approxMatch,
    pureldap.LDAPFilter_present,
    pureldap.LDAPFilter_extensibleMatch,
    ]:
    nevow.flat.registerFlattener(LDAPFilterSerializer, c)

class AddressBookResource(rend.Page):
    addSlash = True

    docFactory = loaders.xmlfile(
        'searchform.xhtml',
        templateDir=os.path.split(os.path.abspath(__file__))[0])

    def configurable_(self, context):
        request = context.locate(inevow.IRequest)
        i = request.session.getComponent(IAddressBookSearch)
        if i is None:
            i = CurrentSearch()
            request.session.setComponent(IAddressBookSearch, i)
        return i

    def data_search(self, context, data):
        d = defer.maybeDeferred(self.locateConfigurable, context, '')
        def cb(configurable):
            return configurable.original
        d.addCallback(cb)
        return d

    def child_form_css(self, request):
        return webform.defaultCSS

    def render_input(self, context, data):
        return webform.renderForms()

    def render_haveSearch(self, context, data):
        r=context.tag.allPatterns(str(data.search is not None))
        return context.tag.clear()[r]

    def render_searchFilter(self, context, data):
        return data.asText()

class AddressBookRealm:
    implements(portal.IRealm)

    def __init__(self, resource):
        self.resource = resource

    def requestAvatar(self, avatarId, mind, *interfaces):
        if inevow.IResource not in interfaces:
            raise NotImplementedError, "no interface"
        return (inevow.IResource,
                self.resource,
                lambda: None)

def getSite(config):
    form = AddressBookResource()
    form.remember(config, ILDAPConfig)
    realm = AddressBookRealm(form)
    site = appserver.NevowSite(
        guard.SessionWrapper(
        portal.Portal(realm, [checkers.AllowAnonymousAccess()])))
    return site
